""" The input data resolution module is a plugin that
    allows to define VO input data policy in a simple way using existing
    utilities in DIRAC or extension code supplied by the VO.

    The arguments dictionary from the Job Wrapper includes the file catalogue
    result and in principle has all the necessary information to resolve input data
    for applications.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import DIRAC
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.ModuleFactory import ModuleFactory
from DIRAC.WorkloadManagementSystem.Client.PoolXMLSlice import PoolXMLSlice
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations

__RCSID__ = "$Id$"

COMPONENT_NAME = "InputDataResolution"
CREATE_CATALOG = False


class InputDataResolution(object):
    """Defines the Input Data Policy"""

    #############################################################################
    def __init__(self, argumentsDict):
        """Standard constructor"""
        self.arguments = argumentsDict
        self.name = COMPONENT_NAME
        self.log = gLogger.getSubLogger(self.name)
        op = Operations()
        self.arguments.setdefault("Configuration", {})["AllReplicas"] = op.getValue(
            "InputDataPolicy/AllReplicas", False
        )
        self.arguments["Configuration"].setdefault("Protocol", op.getValue("InputDataPolicy/Protocols/Local", []))
        self.arguments["Configuration"].setdefault(
            "RemoteProtocol", op.getValue("InputDataPolicy/Protocols/Remote", [])
        )

        # By default put input data into the current directory
        self.arguments.setdefault("InputDataDirectory", "CWD")

    #############################################################################
    def execute(self):
        """Given the arguments from the Job Wrapper, this function calls existing
        utilities in DIRAC to resolve input data.
        """
        resolvedInputData = self.__resolveInputData()
        if not resolvedInputData["OK"]:
            self.log.error("InputData resolution failed with result:\n%s" % (resolvedInputData["Message"]))
            return resolvedInputData

        # For local running of this module we can expose an option to ignore missing files
        ignoreMissing = self.arguments.get("IgnoreMissing", False)

        # Missing some of the input files is a fatal error unless ignoreMissing option is defined
        failedReplicas = resolvedInputData["Value"].get("Failed", {})
        if failedReplicas and not ignoreMissing:
            self.log.error("Failed to obtain access to the following files:\n%s" % ("\n".join(sorted(failedReplicas))))
            return S_ERROR("Failed to access some of requested input data")

        if not resolvedInputData["Value"].get("Successful"):
            return S_ERROR("Could not access any requested input data")

        if CREATE_CATALOG:
            res = self._createCatalog(resolvedInputData)
            if not res["OK"]:
                return res

        return resolvedInputData

    #############################################################################

    def _createCatalog(self, resolvedInputData, catalogName="pool_xml_catalog.xml", pfnType="ROOT_All"):
        """By default uses PoolXMLSlice, VO extensions can modify at will"""

        resolvedData = resolvedInputData["Successful"]
        tmpDict = {}
        for lfn, mdata in resolvedData.items():
            tmpDict[lfn] = mdata
            tmpDict[lfn]["pfntype"] = pfnType
            self.log.verbose("Adding PFN file type %s for LFN:%s" % (pfnType, lfn))

        catalogName = self.arguments["Configuration"].get("CatalogName", catalogName)
        self.log.verbose("Catalog name will be: %s" % catalogName)

        resolvedData = tmpDict
        appCatalog = PoolXMLSlice(catalogName)
        return appCatalog.execute(resolvedData)

    #############################################################################

    def __resolveInputData(self):
        """This method controls the execution of the DIRAC input data modules according
        to the VO policy defined in the configuration service.
        """
        site = self.arguments["Configuration"].get("SiteName", DIRAC.siteName())

        self.arguments.setdefault("Job", {})

        policy = self.arguments["Job"].get("InputDataPolicy", [])
        if policy:
            # In principle this can be a list of modules with the first taking precedence
            if isinstance(policy, six.string_types):
                policy = [policy]
            self.log.info("Job has a specific policy setting: %s" % (", ".join(policy)))
        else:
            self.log.debug("Attempting to resolve input data policy for site %s" % site)
            inputDataPolicy = Operations().getOptionsDict("InputDataPolicy")
            if not inputDataPolicy["OK"]:
                return S_ERROR("Could not resolve InputDataPolicy from Operations InputDataPolicy")

            options = inputDataPolicy["Value"]
            policy = options.get(site, options.get("Default", []))
            if policy:
                policy = [x.strip() for x in policy.split(",")]
                if site in options:
                    prStr = "Found specific"
                else:
                    prStr = "Applying default"
                self.log.info("%s input data policy for site %s:\n%s" % (prStr, site, "\n".join(policy)))

        dataToResolve = []  # if none, all supplied input data is resolved
        successful = {}
        for modulePath in policy:
            result = self.__runModule(modulePath, dataToResolve)
            if not result["OK"]:
                self.log.warn("Problem during %s execution" % modulePath)
                return result

            result = result["Value"]
            successful.update(result.get("Successful", {}))
            dataToResolve = result.get("Failed", [])
            if dataToResolve:
                self.log.info("%s failed for the following files:\n%s" % (modulePath, "\n".join(dataToResolve)))
            else:
                self.log.info("All replicas resolved after %s execution" % (modulePath))
                break

        if successful:
            self.log.verbose("Successfully resolved:", str(successful))

        return S_OK({"Successful": successful, "Failed": dataToResolve})

    #############################################################################
    def __runModule(self, modulePath, remainingReplicas):
        """This method provides a way to run the modules specified by the VO that
        govern the input data access policy for the current site. Using the
        InputDataPolicy section from Operations different modules can be defined for
        particular sites or for InputDataPolicy defined in the JDL of the jobs.
        """
        self.log.info("Attempting to run %s" % (modulePath))
        moduleFactory = ModuleFactory()
        moduleInstance = moduleFactory.getModule(modulePath, self.arguments)
        if not moduleInstance["OK"]:
            return moduleInstance

        module = moduleInstance["Value"]
        result = module.execute(remainingReplicas)
        return result


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



"""
This class a wrapper around elasticsearch-py.
It is used to query Elasticsearch instances.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six

from datetime import datetime
from datetime import timedelta

import certifi
import copy
import functools
import json

from elasticsearch import Elasticsearch
from elasticsearch_dsl import Search, Q, A
from elasticsearch.exceptions import ConnectionError, TransportError, NotFoundError, RequestError
from elasticsearch.helpers import BulkIndexError, bulk

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities import Time, DErrno
from DIRAC.FrameworkSystem.Client.BundleDeliveryClient import BundleDeliveryClient


sLog = gLogger.getSubLogger(__name__)


def ifConnected(method):
    """Decorator for checking that the connection is established."""

    @functools.wraps(method)
    def wrapper_decorator(self, *args, **kwargs):
        if self._connected:
            return method(self, *args, **kwargs)
        else:
            sLog.error("Not connected")
            return S_ERROR("Not connected")

    return wrapper_decorator


def generateDocs(data, withTimeStamp=True):
    """Generator for fast bulk indexing, yields docs

    :param list data: list of dictionaries
    :param bool withTimeStamp: add the timestamps to the docs

    :return: doc
    """
    for doc in copy.deepcopy(data):
        if "_type" not in doc:
            doc["_type"] = "_doc"
        if withTimeStamp:
            if "timestamp" not in doc:
                sLog.warn("timestamp is not given")

            # if the timestamp is not provided, we use the current utc time.
            timestamp = doc.get("timestamp", int(Time.toEpoch()))
            try:
                if isinstance(timestamp, datetime):
                    doc["timestamp"] = int(timestamp.strftime("%s")) * 1000
                elif isinstance(timestamp, six.string_types):
                    timeobj = datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S.%f")
                    doc["timestamp"] = int(timeobj.strftime("%s")) * 1000
                else:  # we assume  the timestamp is an unix epoch time (integer).
                    doc["timestamp"] = timestamp * 1000
            except (TypeError, ValueError) as e:
                # in case we are not able to convert the timestamp to epoch time....
                sLog.error("Wrong timestamp", e)
                doc["timestamp"] = int(Time.toEpoch()) * 1000

        sLog.debug("yielding %s" % doc)
        yield doc


class ElasticSearchDB(object):

    """
    .. class:: ElasticSearchDB

    :param str url: the url to the database for example: el.cern.ch:9200
    :param str gDebugFile: is used to save the debug information to a file
    :param int timeout: the default time out to Elasticsearch
    :param int RESULT_SIZE: The number of data points which will be returned by the query.
    """

    __url = ""
    __timeout = 120
    clusterName = ""
    RESULT_SIZE = 10000

    ########################################################################
    def __init__(
        self,
        host,
        port,
        user=None,
        password=None,
        indexPrefix="",
        useSSL=True,
        useCRT=False,
        ca_certs=None,
        client_key=None,
        client_cert=None,
    ):
        """c'tor

        :param self: self reference
        :param str host: name of the database for example: MonitoringDB
        :param str port: The full name of the database for example: 'Monitoring/MonitoringDB'
        :param str user: user name to access the db
        :param str password: if the db is password protected we need to provide a password
        :param str indexPrefix: it is the indexPrefix used to get all indexes
        :param bool useSSL: We can disable using secure connection. By default we use secure connection.
        :param bool useCRT: Use certificates.
        :param str ca_certs: CA certificates bundle.
        :param str client_key: Client key.
        :param str client_cert: Client certificate.
        """

        self.__indexPrefix = indexPrefix
        self._connected = False
        if user and password:
            sLog.debug("Specified username and password")
            if port:
                self.__url = "https://%s:%s@%s:%d" % (user, password, host, port)
            else:
                self.__url = "https://%s:%s@%s" % (user, password, host)
        else:
            sLog.debug("Username and password not specified")
            if port:
                self.__url = "http://%s:%d" % (host, port)
            else:
                self.__url = "http://%s" % host

        if port:
            sLog.verbose("Connecting to %s:%s, useSSL = %s" % (host, port, useSSL))
        else:
            sLog.verbose("Connecting to %s, useSSL = %s" % (host, useSSL))

        if useSSL:
            if ca_certs:
                casFile = ca_certs
            else:
                bd = BundleDeliveryClient()
                retVal = bd.getCAs()
                casFile = None
                if not retVal["OK"]:
                    sLog.error("CAs file does not exists:", retVal["Message"])
                    casFile = certifi.where()
                else:
                    casFile = retVal["Value"]

            self.client = Elasticsearch(
                self.__url, timeout=self.__timeout, use_ssl=True, verify_certs=True, ca_certs=casFile
            )
        elif useCRT:
            self.client = Elasticsearch(
                self.__url,
                timeout=self.__timeout,
                use_ssl=True,
                verify_certs=True,
                ca_certs=ca_certs,
                client_cert=client_cert,
                client_key=client_key,
            )
        else:
            self.client = Elasticsearch(self.__url, timeout=self.__timeout)

        # Before we use the database we try to connect
        # and retrieve the cluster name

        try:
            if self.client.ping():
                # Returns True if the cluster is running, False otherwise
                result = self.client.info()
                self.clusterName = result.get("cluster_name", " ")  # pylint: disable=no-member
                sLog.info("Database info\n", json.dumps(result, indent=4))
                self._connected = True
            else:
                sLog.error("Cannot ping ElasticsearchDB!")
        except ConnectionError as e:
            sLog.error(repr(e))

    ########################################################################
    def getIndexPrefix(self):
        """
        It returns the DIRAC setup.
        """
        return self.__indexPrefix

    ########################################################################
    @ifConnected
    def query(self, index, query):
        """Executes a query and returns its result (uses ES DSL language).

        :param self: self reference
        :param str index: index name
        :param dict query: It is the query in ElasticSearch DSL language

        """
        try:
            esDSLQueryResult = self.client.search(index=index, body=query)
            return S_OK(esDSLQueryResult)
        except RequestError as re:
            return S_ERROR(re)

    @ifConnected
    def update(self, index, query=None, updateByQuery=True, id=None):
        """Executes an update of a document, and returns S_OK/S_ERROR

        :param self: self reference
        :param str index: index name
        :param dict query: It is the query in ElasticSearch DSL language
        :param bool updateByQuery: A bool to determine update by query or index values using index function.
        :param int id: ID for the document to be created.

        """

        sLog.debug("Updating %s with %s, updateByQuery=%s, id=%s" % (index, query, updateByQuery, id))

        if not index or not query:
            return S_ERROR("Missing index or query")

        try:
            if updateByQuery:
                esDSLQueryResult = self.client.update_by_query(index=index, body=query)
            else:
                esDSLQueryResult = self.client.index(index=index, doc_type="_doc", body=query, id=id)
            return S_OK(esDSLQueryResult)
        except RequestError as re:
            return S_ERROR(re)

    @ifConnected
    def _Search(self, indexname):
        """
        it returns the object which can be used for retreiving certain value from the DB
        """
        return Search(using=self.client, index=indexname)

    ########################################################################
    def _Q(self, name_or_query="match", **params):
        """
        It is a wrapper to ElasticDSL Query module used to create a query object.
        :param str name_or_query is the type of the query
        """
        return Q(name_or_query, **params)

    def _A(self, name_or_agg, aggsfilter=None, **params):
        """
        It is a wrapper to ElasticDSL aggregation module, used to create an aggregation
        """
        return A(name_or_agg, aggsfilter, **params)

    ########################################################################
    @ifConnected
    def getIndexes(self, indexName=None):
        """
        It returns the available indexes...
        """
        if not indexName:
            indexName = self.__indexPrefix
        sLog.debug("Getting indices alias of %s" % indexName)
        # we only return indexes which belong to a specific prefix for example 'lhcb-production' or 'dirac-production etc.
        return list(self.client.indices.get_alias("%s*" % indexName))

    ########################################################################
    @ifConnected
    def getDocTypes(self, indexName):
        """
        Returns mappings, by index.

        :param str indexName: is the name of the index...
        :return: S_OK or S_ERROR
        """
        result = []
        try:
            sLog.debug("Getting mappings for ", indexName)
            result = self.client.indices.get_mapping(indexName)
        except Exception as e:  # pylint: disable=broad-except
            sLog.exception()
            return S_ERROR(e)

        doctype = ""
        for indexConfig in result:
            if not result[indexConfig].get("mappings"):
                # there is a case when the mapping exits and the value is None...
                # this is usually an empty index or a corrupted index.
                sLog.warn("Index does not have mapping %s!" % indexConfig)
                continue
            if result[indexConfig].get("mappings"):
                doctype = result[indexConfig]["mappings"]
                break  # we suppose the mapping of all indexes are the same...

        if not doctype:
            return S_ERROR("%s does not exists!" % indexName)

        return S_OK(doctype)

    ########################################################################
    @ifConnected
    def existingIndex(self, indexName):
        """
        Checks the existance of an index, by its name

        :param str indexName: the name of the index
        :returns: S_OK/S_ERROR if the request is successful
        """
        sLog.debug("Checking existance of index %s" % indexName)
        try:
            return S_OK(self.client.indices.exists(indexName))
        except TransportError as e:
            sLog.exception()
            return S_ERROR(e)

    ########################################################################

    @ifConnected
    def createIndex(self, indexPrefix, mapping=None, period="day"):
        """
        :param str indexPrefix: it is the index name.
        :param dict mapping: the configuration of the index.
        :param str period: We can specify, which kind of index will be created.
                           Currently only daily and monthly indexes are supported.

        """
        if period is not None:
            fullIndex = self.generateFullIndexName(indexPrefix, period)  # we have to create an index each period...
        else:
            sLog.warn("The period is not provided, so using non-periodic indexes names")
            fullIndex = indexPrefix

        res = self.existingIndex(fullIndex)
        if not res["OK"]:
            return res
        elif res["Value"]:
            return S_OK(fullIndex)

        try:
            sLog.info("Create index: ", fullIndex + str(mapping))
            try:
                self.client.indices.create(index=fullIndex, body={"mappings": mapping})  # ES7
            except RequestError as re:
                if re.error == "mapper_parsing_exception":
                    self.client.indices.create(index=fullIndex, body={"mappings": {"_doc": mapping}})  # ES6
            return S_OK(fullIndex)
        except Exception as e:  # pylint: disable=broad-except
            sLog.error("Can not create the index:", repr(e))
            return S_ERROR("Can not create the index")

    @ifConnected
    def deleteIndex(self, indexName):
        """
        :param str indexName: the name of the index to be deleted...
        """
        sLog.info("Deleting index", indexName)
        try:
            retVal = self.client.indices.delete(indexName)
        except NotFoundError:
            sLog.warn("Index does not exist", indexName)
            return S_OK("Noting to delete")
        except ValueError as e:
            return S_ERROR(DErrno.EVALUE, e)

        if retVal.get("acknowledged"):
            # if the value exists and the value is not None
            sLog.info("Deleted index", indexName)
            return S_OK(indexName)

        return S_ERROR(retVal)

    def index(self, indexName, body=None, docID=None):
        """
        :param str indexName: the name of the index to be used
        :param dict body: the data which will be indexed (basically the JSON)
        :param int id: optional document id
        :return: the index name in case of success.
        """

        sLog.debug("Indexing in %s body %s, id=%s" % (indexName, body, docID))

        if not indexName or not body:
            return S_ERROR("Missing index or body")

        try:
            res = self.client.index(index=indexName, doc_type="_doc", body=body, id=docID)
        except (RequestError, TransportError) as e:
            sLog.exception()
            return S_ERROR(e)

        if res.get("created") or res.get("result") in ("created", "updated"):
            # the created index exists but the value can be None.
            return S_OK(indexName)

        return S_ERROR(res)

    @ifConnected
    def bulk_index(self, indexPrefix, data=None, mapping=None, period="day", withTimeStamp=True):
        """
        :param str indexPrefix: index name.
        :param list data: contains a list of dictionary
        :param dict mapping: the mapping used by elasticsearch
        :param str period: Accepts 'day' and 'month'. We can specify which kind of indexes will be created.
        :param bool withTimeStamp: add timestamp to data, if not there already.

        :returns: S_OK/S_ERROR
        """
        sLog.verbose("Bulk indexing", "%d records will be inserted" % len(data))
        if mapping is None:
            mapping = {}

        if period is not None:
            indexName = self.generateFullIndexName(indexPrefix, period)
        else:
            indexName = indexPrefix
        sLog.debug("Bulk indexing into %s of %s" % (indexName, data))

        res = self.existingIndex(indexName)
        if not res["OK"]:
            return res
        if not res["Value"]:
            retVal = self.createIndex(indexPrefix, mapping, period)
            if not retVal["OK"]:
                return retVal

        try:
            res = bulk(client=self.client, index=indexName, actions=generateDocs(data, withTimeStamp))
        except (BulkIndexError, RequestError) as e:
            sLog.exception()
            return S_ERROR(e)

        if res[0] == len(data):
            # we have inserted all documents...
            return S_OK(len(data))
        else:
            return S_ERROR(res)

    @ifConnected
    def getUniqueValue(self, indexName, key, orderBy=False):
        """
        :param str indexName: the name of the index which will be used for the query
        :param dict orderBy: it is a dictionary in case we want to order the result {key:'desc'} or {key:'asc'}
        :returns: a list of unique value for a certain key from the dictionary.
        """

        query = self._Search(indexName)

        endDate = datetime.utcnow()

        startDate = endDate - timedelta(days=30)

        timeFilter = self._Q(
            "range",
            timestamp={
                "lte": int(Time.toEpoch(endDate)) * 1000,
                "gte": int(Time.toEpoch(startDate)) * 1000,
            },
        )
        query = query.filter("bool", must=timeFilter)
        if orderBy:
            query.aggs.bucket(key, "terms", field=key, size=self.RESULT_SIZE, order=orderBy).metric(
                key, "cardinality", field=key
            )
        else:
            query.aggs.bucket(key, "terms", field=key, size=self.RESULT_SIZE).metric(key, "cardinality", field=key)

        try:
            query = query.extra(size=self.RESULT_SIZE)  # do not need the raw data.
            sLog.debug("Query", query.to_dict())
            result = query.execute()
        except TransportError as e:
            return S_ERROR(e)

        values = []
        for bucket in result.aggregations[key].buckets:
            values += [bucket["key"]]
        del query
        sLog.debug("Nb of unique rows retrieved", len(values))
        return S_OK(values)

    def pingDB(self):
        """
        Try to connect to the database

        :return: S_OK(TRUE/FALSE)
        """
        connected = False
        try:
            connected = self.client.ping()
        except ConnectionError as e:
            sLog.error("Cannot connect to the db", repr(e))
        return S_OK(connected)

    @ifConnected
    def deleteByQuery(self, indexName, query):
        """
        Delete data by query (careful!)

        :param str indexName: the name of the index
        :param str query: the JSON-formatted query for which we want to issue the delete
        """
        try:
            self.client.delete_by_query(index=indexName, body=query)
        except Exception as inst:
            sLog.error("ERROR: Couldn't delete data")
            return S_ERROR(inst)
        return S_OK("Successfully deleted data from index %s" % indexName)

    @staticmethod
    def generateFullIndexName(indexName, period):
        """
        Given an index prefix we create the actual index name.

        :param str indexName: it is the name of the index
        :param str period: We can specify which kind of indexes will be created (day, week, month, year, null).
        :returns: string with full index name
        """

        # if the period is not correct, we use no-period indexes (same as "null").
        if period.lower() not in ["day", "week", "month", "year", "null"]:
            sLog.error("Period is not correct: ", period)
            return indexName
        elif period.lower() == "day":
            today = datetime.today().strftime("%Y-%m-%d")
            return "%s-%s" % (indexName, today)
        elif period.lower() == "week":
            week = datetime.today().isocalendar()[1]
            return "%s-%s" % (indexName, week)
        elif period.lower() == "month":
            month = datetime.today().strftime("%Y-%m")
            return "%s-%s" % (indexName, month)
        elif period.lower() == "year":
            year = datetime.today().strftime("%Y")
            return "%s-%s" % (indexName, year)
        elif period.lower() == "null":
            return indexName



""" :mod: RucioSynchronizer

  Agent that synchronizes Rucio and Dirac

The following options can be set for the RucioSynchronizerAgent.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN RucioSynchronizerAgent
  :end-before: ##END
  :dedent: 2
  :caption: RucioSynchronizerAgent options
"""

# # imports
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from itertools import permutations
from traceback import format_exc

from rucio.client import Client
from rucio.common.exception import RSEProtocolNotSupported, Duplicate, RSEAttributeNotFound

from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import (
    getUserOption,
    getAllUsers,
    getHosts,
    getHostOption,
    getAllGroups,
    getDNsInGroup,
)
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers, resolveSEGroup
from DIRAC.Resources.Storage.StorageElement import StorageElement


__RCSID__ = "Id$"


def getStorageElements():
    """
    Get configuration of storage elements

    :return: S_OK/S_ERROR, Value dictionary with key SE and value protocol list
    """
    seProtocols = {}
    dms = DMSHelpers()
    for seName in dms.getStorageElements():
        seProtocols[seName] = []
        se = StorageElement(seName)
        all_protocols = []

        read_protocols = {}
        protocols = se.options.get("AccessProtocols")
        if not protocols:
            continue
        idx = 1
        for prot in protocols:
            read_protocols[prot] = idx
            idx += 1
            if prot not in all_protocols:
                all_protocols.append(prot)
        write_protocols = {}
        protocols = se.options.get("WriteProtocols")
        if not protocols:
            continue
        idx = 1
        for prot in protocols:
            write_protocols[prot] = idx
            idx += 1
            if prot not in all_protocols:
                all_protocols.append(prot)

        mapping = {"Protocol": "scheme", "Host": "hostname", "Port": "port", "Path": "prefix"}
        for protocol in all_protocols:
            space_token = None
            params = {
                "hostname": None,
                "scheme": None,
                "port": None,
                "prefix": None,
                "impl": "rucio.rse.protocols.gfal.Default",
                "domains": {
                    "lan": {"read": 0, "write": 0, "delete": 0},
                    "wan": {"read": 0, "write": 0, "delete": 0, "third_party_copy": 0},
                },
            }
            res = se.getStorageParameters(protocol=protocol)
            if res["OK"]:
                values = res["Value"]
                for key in ["Protocol", "Host", "Access", "Path", "Port", "WSUrl", "SpaceToken", "WSUrl", "PluginName"]:
                    value = values.get(key)
                    if key in mapping:
                        params[mapping[key]] = value
                    else:
                        if key == "SpaceToken":
                            space_token = value
                        if params["scheme"] == "srm" and key == "WSUrl":
                            params["extended_attributes"] = {
                                "web_service_path": "%s" % value,
                                "space_token": space_token,
                            }
                    if key == "Protocol":
                        params["domains"]["lan"]["read"] = read_protocols.get(value, 0)
                        params["domains"]["wan"]["read"] = read_protocols.get(value, 0)
                        params["domains"]["lan"]["write"] = write_protocols.get(value, 0)
                        params["domains"]["wan"]["write"] = write_protocols.get(value, 0)
                        params["domains"]["lan"]["delete"] = write_protocols.get(value, 0)
                        params["domains"]["wan"]["delete"] = write_protocols.get(value, 0)
                        params["domains"]["wan"]["third_party_copy"] = write_protocols.get(value, 0)
                seProtocols[seName].append(params)
    return S_OK(seProtocols)


class RucioSynchronizerAgent(AgentModule):
    """
    .. class::  RucioSynchronizerAgent

    Agent that synchronizes Rucio and Dirac
    """

    def execute(self):
        """execution in one agent's cycle

        :param self: self reference
        """

        valid_protocols = ["srm", "gsiftp", "davs", "https", "root"]
        default_email = None
        try:
            client = Client(account="root", auth_type="userpass")

            # Get the storage elements from Dirac Configuration and create them in Rucio
            newRSE = False
            self.log.info("Synchronizing SEs")
            result = getStorageElements()
            if result["OK"]:
                rses = [rse["rse"] for rse in client.list_rses()]
                for se in result["Value"]:
                    if se not in rses:
                        # The SE doesn't exist. Will create it
                        newRSE = True
                        self.log.info("%s does not exist and will be created" % se)
                        try:
                            client.add_rse(rse=se, deterministic=True, volatile=False)
                        except Exception as err:
                            self.log.error("Cannot create RSE" % str(err))
                            continue

                        # Add RSE attributes for the new RSE
                        ret = gConfig.getOptionsDict("Resources/FTSEndpoints/FTS3")
                        ftsList = ""
                        if ret["OK"]:
                            ftsList = ",".join(ret["Value"].values())
                        dictRSEAttributes = {"naming_convention": "BelleII", "ANY": True, "fts": ftsList}
                        for key in dictRSEAttributes:
                            self.log.info("On %s, setting %s : %s", se, key, dictRSEAttributes[key])
                            client.add_rse_attribute(se, key, value=dictRSEAttributes[key])
                        client.set_local_account_limit("root", se, 100000000000000000)

                    # Create the protocols
                    try:
                        protocols = client.get_protocols(se)
                    except RSEProtocolNotSupported as err:
                        self.log.info("Cannot get protocols for %s : %s", se, str(err))
                        protocols = []
                    existing_protocols = []
                    for prot in protocols:
                        existing_protocols.append(
                            (str(prot["scheme"]), str(prot["hostname"]), str(prot["port"]), str(prot["prefix"]))
                        )
                    protocols_to_create = []
                    for params in result["Value"][se]:
                        prot = (
                            str(params["scheme"]),
                            str(params["hostname"]),
                            str(params["port"]),
                            str(params["prefix"]),
                        )
                        protocols_to_create.append(prot)
                        if prot not in existing_protocols and prot[0] in valid_protocols:
                            # The protocol defined in Dirac does not exist in Rucio. Will be created
                            self.log.info(
                                "Will create new protocol %s://%s:%s%s on %s",
                                params["scheme"],
                                params["hostname"],
                                params["port"],
                                params["prefix"],
                                se,
                            )
                            try:
                                client.add_protocol(rse=se, params=params)
                            except Duplicate:
                                self.log.info("Protocol %s already exists on %s", params["scheme"], se)
                            except Exception as err:
                                self.log.error("Cannot create protocol on RSE %s : %s", se, str(err))
                        else:
                            update = False
                            for protocol in protocols:
                                if prot == (
                                    str(protocol["scheme"]),
                                    str(protocol["hostname"]),
                                    str(protocol["port"]),
                                    str(protocol["prefix"]),
                                ):
                                    # Check if the protocol defined in Dirac has the same priority as the one defined in Rucio
                                    for domain in ["lan", "wan"]:
                                        for activity in ["read", "write", "delete"]:
                                            if (
                                                params["domains"][domain][activity]
                                                != protocol["domains"][domain][activity]
                                            ):
                                                update = True
                                                break

                                    if (
                                        params["domains"]["wan"]["third_party_copy"]
                                        != protocol["domains"]["wan"]["third_party_copy"]
                                    ):
                                        update = True
                                    if update:
                                        data = {
                                            "prefix": params["prefix"],
                                            "read_lan": params["domains"]["lan"]["read"],
                                            "read_wan": params["domains"]["wan"]["read"],
                                            "write_lan": params["domains"]["lan"]["write"],
                                            "write_wan": params["domains"]["wan"]["write"],
                                            "delete_lan": params["domains"]["lan"]["delete"],
                                            "delete_wan": params["domains"]["wan"]["delete"],
                                            "third_party_copy": params["domains"]["wan"]["write"],
                                        }
                                        self.log.info(
                                            "Will update protocol %s://%s:%s%s on %s",
                                            params["scheme"],
                                            params["hostname"],
                                            params["port"],
                                            params["prefix"],
                                            se,
                                        )
                                        client.update_protocols(
                                            rse=se,
                                            scheme=params["scheme"],
                                            data=data,
                                            hostname=params["hostname"],
                                            port=params["port"],
                                        )
                    for prot in existing_protocols:
                        if prot not in protocols_to_create:
                            self.log.info(
                                "Will delete protocol %s://%s:%s%s on %s", prot[0], prot[1], prot[2], prot[3], se
                            )
                            client.delete_protocols(se, scheme=prot[0], hostname=prot[1], port=prot[2])
            else:
                self.log.error("Cannot get SEs : %s", result["Value"])

            # If new RSE added, add distances
            rses = [rse["rse"] for rse in client.list_rses()]
            if newRSE:
                self.log.info("Adding distances")
                for src_rse, dest_rse in permutations(rses, r=2):
                    try:
                        client.add_distance(src_rse, dest_rse, {"ranking": 1, "distance": 10})
                    except Exception as err:
                        self.log.error("Cannot add distance for %s:%s : %s", src_rse, dest_rse, str(err))

            # Collect the shares from Dirac Configuration and create them in Rucio
            self.log.info("Synchronizing shares")
            result = Operations().getOptionsDict("Production/SEshares")
            if result["OK"]:
                rseDict = result["Value"]
                for rse in rses:
                    try:
                        self.log.info("Setting productionSEshare for %s : %s", rse, rseDict.get(rse, 0))
                        client.add_rse_attribute(rse, "productionSEshare", rseDict.get(rse, 0))
                    except Exception:
                        self.log.error("Cannot create productionSEshare for %s", rse)
            else:
                self.log.error("Cannot get SEs : %s", result["Value"])

            result = Operations().getSections("Shares")
            if result["OK"]:
                for dataLevel in result["Value"]:
                    result = Operations().getOptionsDict("Shares/%s" % dataLevel)
                    if not result["OK"]:
                        self.log.error("Cannot get SEs : %s", result["Value"])
                        continue
                    rseDict = result["Value"]
                    for rse in rses:
                        try:
                            self.log.info("Setting %sShare for %s : %s", dataLevel, rse, rseDict.get(rse, 0))
                            client.add_rse_attribute(rse, "%sShare" % dataLevel, rseDict.get(rse, 0))
                        except Exception:
                            self.log.error("Cannot create %sShare for %s", dataLevel, rse)
            else:
                self.log.error("Cannot get shares : %s", result["Value"])

            # Create the RSE attribute PrimaryDataSE and OccupancyLFN
            result = gConfig.getValue("Resources/StorageElementGroups/PrimarySEs")
            result = getStorageElements()
            if result["OK"]:
                allSEs = result["Value"]
                primarySEs = resolveSEGroup("PrimarySEs", allSEs)
                self.log.info("Will set primarySEs flag to %s", str(primarySEs))
                for rse in rses:
                    if rse in allSEs:
                        occupancyLFN = StorageElement(rse).options.get("OccupancyLFN")
                        try:
                            client.add_rse_attribute(rse, "OccupancyLFN", occupancyLFN)
                        except Exception as err:
                            self.log.error("Cannot create RSE attribute OccupancyLFN for %s : %s", rse, str(err))
                    if rse in primarySEs:
                        try:
                            client.add_rse_attribute(rse, "PrimaryDataSE", True)
                        except Exception as err:
                            self.log.error("Cannot create RSE attribute PrimaryDataSE for %s : %s", rse, str(err))
                    else:
                        try:
                            client.delete_rse_attribute(rse, "PrimaryDataSE")
                        except RSEAttributeNotFound:
                            pass
                        except Exception as err:
                            self.log.error("Cannot remove RSE attribute PrimaryDataSE for %s : %s", rse, str(err))

            # Collect the user accounts from Dirac Configuration and create user accounts in Rucio
            self.log.info("Synchronizing accounts")
            listAccounts = [str(acc["account"]) for acc in client.list_accounts()]
            listScopes = [str(scope) for scope in client.list_scopes()]
            dnMapping = {}
            for account in getAllUsers():
                dn = getUserOption(account, "DN")
                email = getUserOption(account, "Email")
                dnMapping[dn] = email
                if account not in listAccounts:
                    self.log.info("Will create %s with associated DN %s", account, dn)
                    try:
                        client.add_account(account, "USER", email)
                        listAccounts.append(account)
                    except Exception as err:
                        self.log.error("Cannot create account %s : %s", account, str(err))
                    try:
                        client.add_identity(account=account, identity=dn, authtype="X509", email=email, default=True)
                    except Exception as err:
                        self.log.error("Cannot create account %s : %s", account, str(err))
                    for rse in rses:
                        client.set_local_account_limit(account, rse, 1000000000000000)
                else:
                    try:
                        client.add_identity(account=account, identity=dn, authtype="X509", email=email, default=True)
                    except Duplicate:
                        pass
                    except Exception as err:
                        self.log.error("Cannot create identity %s for account %s : %s", dn, account, str(err))
                scope = "user." + account
                if scope not in listScopes:
                    try:
                        self.log.info("Will create scope %s", scope)
                        client.add_scope(account, scope)
                        self.log.info("Scope %s successfully added", scope)
                    except Exception as err:
                        self.log.error("Cannot create scope %s : %s", scope, str(err))

            # Collect the group accounts from Dirac Configuration and create service accounts in Rucio
            groups = getAllGroups()
            for group in groups:
                if group not in listAccounts:
                    self.log.info("Will create SERVICE account %s" % (group))
                    try:
                        client.add_account(group, "SERVICE", None)
                        listAccounts.append(group)
                    except Exception as err:
                        self.log.error("Cannot create account %s : %s" % (group, str(err)))
                    for rse in rses:
                        client.set_local_account_limit(account, rse, 1000000000000000)

                for dn in getDNsInGroup(group):
                    try:
                        client.add_identity(
                            account=group, identity=dn, authtype="X509", email=dnMapping.get(dn, default_email)
                        )
                    except Duplicate:
                        pass
                    except Exception as err:
                        self.log.error("Cannot create identity %s for account %s : %s" % (dn, group, str(err)))
                        self.log.error(format_exc())

            # Collect the group accounts from Dirac Configuration and create service accounts in Rucio
            result = getHosts()
            if not result["OK"]:
                self.log.error("Cannot get host accounts : %s" % result["Value"])
            else:
                hosts = result["Value"]
                for host in hosts:
                    dn = getHostOption(host, "DN")
                    email = dnMapping.get(dn, default_email)
                    try:
                        client.add_identity(account="dirac_srv", identity=dn, authtype="X509", email=email)
                    except Duplicate:
                        pass
                    except Exception as err:
                        self.log.error("Cannot create identity %s for account dirac_srv : %s" % (dn, str(err)))
                        self.log.error(format_exc())

            return S_OK()
        except Exception:
            return S_ERROR(str(format_exc()))



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
import threading
from DIRAC import gLogger, S_ERROR
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler


class TransportPool(object):
    def __init__(self, logger=False):
        if logger:
            self.log = logger
        else:
            self.log = gLogger
        self.__modLock = threading.Lock()
        self.__transports = {}
        self.__listenPersistConn = False
        self.__msgCounter = 0
        result = gThreadScheduler.addPeriodicTask(5, self.__sendKeepAlives)
        if not result["OK"]:
            self.log.fatal("Cannot add task to thread scheduler", result["Message"])
        self.__keepAlivesTask = result["Value"]

    #
    # Send keep alives
    #

    def __sendKeepAlives(self, retries=5):
        if retries == 0:
            return
        tridList = []
        try:
            tridList = [trid for trid in self.__transports]
        except RuntimeError:
            self.__sendKeepAlives(retries - 1)
        for trid in tridList:
            try:
                tr = self.__transports[trid][0]
            except KeyError:
                continue
            if not tr.getKeepAliveLapse():
                continue
            try:
                tr.sendKeepAlive(now=time.time())
            except KeyError:
                continue
            except Exception:
                gLogger.exception("Cannot send keep alive")

    # exists

    def exists(self, trid):
        return trid in self.__transports

    # Add

    def add(self, transport):
        remoteAddr = transport.getRemoteAddress()
        localAddr = transport.getLocalAddress()
        self.log.debug("New connection -> %s:%s" % (remoteAddr[0], remoteAddr[1]))
        trid = "%s:%s->%s:%s" % (localAddr[0], localAddr[1], remoteAddr[0], remoteAddr[1])
        return self.__add(trid, transport)

    def __add(self, trid, transport):
        self.__modLock.acquire()
        try:
            if not self.exists(trid):
                self.__transports[trid] = (transport, {})
        finally:
            self.__modLock.release()
        return trid

    # Data association

    def associateData(self, trid, kw, value):
        self.__modLock.acquire()
        try:
            if trid in self.__transports:
                self.__transports[trid][1][kw] = value
        finally:
            self.__modLock.release()

    def getAssociatedData(self, trid, kw):
        try:
            return self.__transports[trid][1][kw]
        except KeyError:
            return None

    # Get transport

    def get(self, trid):
        try:
            return self.__transports[trid][0]
        except KeyError:
            return None

    # Receive
    def receive(self, trid, maxBufferSize=0, blockAfterKeepAlive=True, idleReceive=False):
        try:
            received = self.__transports[trid][0].receiveData(maxBufferSize, blockAfterKeepAlive, idleReceive)
            return received
        except KeyError:
            return S_ERROR("No transport with id %s defined" % trid)

    # Send
    def send(self, trid, msg):
        try:
            transport = self.__transports[trid][0]
        except KeyError:
            return S_ERROR("No transport with id %s defined" % trid)
        return transport.sendData(msg)

    # Send And Close

    def sendErrorAndClose(self, trid, msg):
        try:
            result = self.__transports[trid][0].sendData(S_ERROR(msg))
            if not result["OK"]:
                return result
        except KeyError:
            return S_ERROR("No transport with id %s defined" % trid)
        finally:
            self.close(trid)

    def sendAndClose(self, trid, msg):
        try:
            result = self.__transports[trid][0].sendData(msg)
            if not result["OK"]:
                return result
        except KeyError:
            return S_ERROR("No transport with id %s defined" % trid)
        finally:
            self.close(trid)

    def sendKeepAlive(self, trid, responseId=None):
        try:
            return self.__transports[trid][0].sendKeepAlive(responseId)
        except KeyError:
            return S_ERROR("No transport with id %s defined" % trid)

    # Close

    def close(self, trid):
        try:
            self.__transports[trid][0].close()
        except KeyError:
            return S_ERROR("No transport with id %s defined" % trid)
        self.remove(trid)

    def remove(self, trid):
        self.__modLock.acquire()
        try:
            if trid in self.__transports:
                del self.__transports[trid]
        finally:
            self.__modLock.release()


gTransportPool = None


def getGlobalTransportPool():
    global gTransportPool
    if not gTransportPool:
        gTransportPool = TransportPool()
    return gTransportPool



""" DIRAC FileCatalog utilities
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities.List import intListToString


def getIDSelectString(ids):
    """
    :param ids: input IDs - can be single int, list or tuple or a SELECT string
    :return: Select string
    """
    if isinstance(ids, six.string_types) and ids.lower().startswith("select"):
        idString = ids
    elif isinstance(ids, six.integer_types):
        idString = "%d" % ids
    elif isinstance(ids, (tuple, list)):
        idString = intListToString(ids)
    else:
        return S_ERROR("Illegal fileID")

    return S_OK(idString)



########################################################################
# File :   ComputingElementFactory.py
# Author : Stuart Paterson
########################################################################

"""  The Computing Element Factory has one method that instantiates a given Computing Element
     from the CEUnique ID specified in the JobAgent configuration section.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Resources.Computing.ComputingElement import getCEConfigDict
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader

__RCSID__ = "$Id$"


class ComputingElementFactory(object):

    #############################################################################
    def __init__(self, ceType=""):
        """Standard constructor"""
        self.ceType = ceType
        self.log = gLogger.getSubLogger("ComputingElementFactory")

    #############################################################################
    def getCE(self, ceType="", ceName="", ceParametersDict={}):
        """This method returns the CE instance corresponding to the supplied
        CEUniqueID.  If no corresponding CE is available, this is indicated.
        """
        if ceType:
            self.log.verbose("Creating CE of type %s" % ceType)
        if ceName:
            self.log.verbose("Creating CE for name %s" % ceName)
        ceTypeLocal = ceType if ceType else self.ceType
        ceNameLocal = ceName if ceName else ceType
        ceConfigDict = getCEConfigDict(ceNameLocal)
        self.log.verbose("CEConfigDict", ceConfigDict)
        if "CEType" in ceConfigDict:
            ceTypeLocal = ceConfigDict["CEType"]
        if not ceTypeLocal:
            error = "Can not determine CE Type"
            self.log.error(error)
            return S_ERROR(error)
        subClassName = "%sComputingElement" % (ceTypeLocal)

        result = ObjectLoader().loadObject("Resources.Computing.%s" % subClassName)
        if not result["OK"]:
            self.log.error("Failed to load object", "%s: %s" % (subClassName, result["Message"]))
            return result

        ceClass = result["Value"]
        try:
            computingElement = ceClass(ceNameLocal)
            # Always set the CEType parameter according to instantiated class
            ceDict = {"CEType": ceTypeLocal}
            if ceParametersDict:
                ceDict.update(ceParametersDict)
            result = computingElement.setParameters(ceDict)
            if not result["OK"]:
                return result

        except Exception as x:
            msg = "ComputingElementFactory could not instantiate %s object" % subClassName
            self.log.exception()
            self.log.warn(msg, repr(x))
            return S_ERROR(repr(x))

        return S_OK(computingElement)



""" Cache

This module provides a generic Cache extended to be used on RSS, RSSCache.
This cache features a lazy update method. It will only be updated if it is
empty and there is a new query. If not, it will remain in its previous state.
However, Cache class internal cache: DictCache sets a validity to its entries.
After that, the cache is empty.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import itertools
import random

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.DictCache import DictCache
from DIRAC.Core.Utilities.LockRing import LockRing
from DIRAC.ResourceStatusSystem.Utilities.RssConfiguration import RssConfiguration


class Cache(object):
    """
    Cache basic class.

    WARNING: None of its methods is thread safe. Acquire / Release lock when
    using them !
    """

    def __init__(self, lifeTime, updateFunc):
        """
        Constructor

        :Parameters:
          **lifeTime** - `int`
            Lifetime of the elements in the cache ( seconds ! )
          **updateFunc** - `function`
            This function MUST return a S_OK | S_ERROR object. In the case of the first,
            its value must be a dictionary.

        """

        # We set a 20% of the lifetime randomly, so that if we have thousands of jobs
        # starting at the same time, all the caches will not end at the same time.
        randomLifeTimeBias = 0.2 * random.random()

        self.log = gLogger.getSubLogger(self.__class__.__name__)

        self.__lifeTime = int(lifeTime * (1 + randomLifeTimeBias))
        self.__updateFunc = updateFunc
        # The records returned from the cache must be valid at least 30 seconds.
        self.__validSeconds = 30

        # Cache
        self.__cache = DictCache()
        self.__cacheLock = LockRing()
        self.__cacheLock.getLock(self.__class__.__name__)

    # internal cache object getter

    def cacheKeys(self):
        """
        Cache keys getter

        :returns: list with keys in the cache valid for at least twice the validity period of the element
        """

        # Here we need to have more than the validity period because of the logic of the matching:
        # * get all the keys with validity T
        # * for each key K, get the element K with validity T
        # This logic fails for elements just at the limit of the required time
        return self.__cache.getKeys(validSeconds=self.__validSeconds * 2)

    # acquire / release Locks

    def acquireLock(self):
        """
        Acquires Cache lock
        """
        self.__cacheLock.acquire(self.__class__.__name__)

    def releaseLock(self):
        """
        Releases Cache lock
        """
        self.__cacheLock.release(self.__class__.__name__)

    # Cache getters

    def get(self, cacheKeys):
        """
        Gets values for cacheKeys given, if all are found ( present on the cache and
        valid ), returns S_OK with the results. If any is not neither present not
        valid, returns S_ERROR.

        :Parameters:
          **cacheKeys** - `list`
            list of keys to be extracted from the cache

        :return: S_OK | S_ERROR
        """

        result = {}

        for cacheKey in cacheKeys:
            cacheRow = self.__cache.get(cacheKey, validSeconds=self.__validSeconds)

            if not cacheRow:
                return S_ERROR("Cannot get %s" % str(cacheKey))
            result.update({cacheKey: cacheRow})

        return S_OK(result)

    def check(self, cacheKeys, vO):
        """
        Modified get() method. Attempts to find keys with a vO value appended or 'all'
        value appended. The cacheKeys passed in are 'flattened' cache keys (no vO)
        Gets values for cacheKeys given, if all are found ( present on the cache and
        valid ), returns S_OK with the results. If any is not neither present not
        valid, returns S_ERROR.

        :Parameters:
          **cacheKeys** - `list`
            list of keys to be extracted from the cache

        :return: S_OK | S_ERROR
        """

        result = {}

        for cacheKey in cacheKeys:
            longCacheKey = cacheKey + ("all",)
            cacheRow = self.__cache.get(longCacheKey, validSeconds=self.__validSeconds)
            if not cacheRow:
                longCacheKey = cacheKey + (vO,)
                cacheRow = self.__cache.get(longCacheKey, validSeconds=self.__validSeconds)
                if not cacheRow:
                    return S_ERROR(
                        'Cannot get extended %s (neither for VO = %s nor for "all" Vos)' % (str(cacheKey), vO)
                    )
            result.update({longCacheKey: cacheRow})

        return S_OK(result)

    # Cache refreshers

    def refreshCache(self):
        """
        Purges the cache and gets fresh data from the update function.

        :return: S_OK | S_ERROR. If the first, its content is the new cache.
        """

        self.log.verbose("refreshing...")

        self.__cache.purgeAll()

        newCache = self.__updateFunc()
        if not newCache["OK"]:
            self.log.error(newCache["Message"])
            return newCache

        newCache = self.__updateCache(newCache["Value"])

        self.log.verbose("refreshed")

        return newCache

    # Private methods

    def __updateCache(self, newCache):
        """
        Given the new cache dictionary, updates the internal cache with it. It sets
        a duration to the entries of <self.__lifeTime> seconds.

        :Parameters:
          **newCache** - `dict`
            dictionary containing a new cache

        :return: dictionary. It is newCache argument.
        """

        for cacheKey, cacheValue in newCache.items():
            self.__cache.add(cacheKey, self.__lifeTime, value=cacheValue)

        # We are assuming nothing will fail while inserting in the cache. There is
        # no apparent reason to suspect from that piece of code.
        return S_OK(newCache)


class RSSCache(Cache):
    """
    The RSSCache is an extension of Cache in which the cache keys are pairs of the
    form: ( elementName, statusType ).

    When instantiating one object of RSSCache, we need to specify the RSS elementType
    it applies, e.g. : StorageElement, CE, Queue, ...

    It provides a unique public method `match` which is thread safe. All other
    methods are not !!
    """

    def __init__(self, lifeTime, updateFunc):
        """
        Constructor

        :Parameters:
          **elementType** - `string`
            RSS elementType, e.g.: StorageElement, CE, Queue... note that one RSSCache
            can only hold elements of a single elementType to avoid issues while doing
            the Cartesian product.
          **lifeTime** - `int`
            Lifetime of the elements in the cache ( seconds ! )
          **updateFunc** - `function`
            This function MUST return a S_OK | S_ERROR object. In the case of the first,
            its value must follow the dict format: ( key, value ) being key ( elementName,
            statusType ) and value status.

        """

        super(RSSCache, self).__init__(lifeTime, updateFunc)

        self.allStatusTypes = RssConfiguration().getConfigStatusType()

    def match(self, elementNames, elementType, statusTypes, vO):
        """
        In first instance, if the cache is invalid, it will request a new one from
        the server.
        It make the Cartesian product of elementNames x statusTypes to generate a key
        set that will be compared against the cache set. If the first is included in
        the second, we have a positive match and a dictionary will be returned. Otherwise,
        we have a cache miss.

        However, arguments ( elementNames or statusTypes ) can have a None value. If
        that is the case, they are considered wildcards.

        :Parameters:
          **elementNames** - [ None, `string`, `list` ]
            name(s) of the elements to be matched
          **elementType** - [ `string` ]
            type of the elements to be matched
          **statusTypes** - [ None, `string`, `list` ]
            name(s) of the statusTypes to be matched

        :return: S_OK() || S_ERROR()
        """

        self.acquireLock()
        try:
            return self._match(elementNames, elementType, statusTypes, vO)
        finally:
            # Release lock, no matter what !
            self.releaseLock()

    # Private methods: NOT THREAD SAFE !!

    def _match(self, elementNames, elementType, statusTypes, vO):
        """
        Method doing the actual work. It must be wrapped around locks to ensure no
        disaster happens.

        :Parameters:
          **elementNames** - [ None, `string`, `list` ]
            name(s) of the elements to be matched
          **elementType** - [ `string` ]
            type of the elements to be matched
          **statusTypes** - [ None, `string`, `list` ]
            name(s) of the statusTypes to be matched

        :return: S_OK() || S_ERROR()
        """

        # Gets the entire cache or a new one if it is empty / invalid
        validCache = self.__getValidCache()
        if not validCache["OK"]:
            return validCache
        validCache = validCache["Value"]

        # Gets matched keys
        try:
            matchKeys = self.__match(validCache, elementNames, elementType, statusTypes, vO)
        except IndexError:
            return S_ERROR("RSS cache empty?")

        if not matchKeys["OK"]:
            return matchKeys

        # Gets objects for matched keys. It will return S_ERROR if the cache value
        # has expired in between. It has 30 valid seconds, which means something was
        # extremely slow above.
        if matchKeys["CheckVO"]:
            cacheMatches = self.check(matchKeys["Value"], vO)  # add an appropriate VO to the keys
        else:
            cacheMatches = self.get(matchKeys["Value"])
        if not cacheMatches["OK"]:
            return cacheMatches

        cacheMatches = cacheMatches["Value"]
        if not cacheMatches:
            return S_ERROR("Empty cache for: %s, %s" % (elementNames, elementType))

        # We undo the key into <elementName> and <statusType>
        try:
            cacheMatchesDict = self.__getDictFromCacheMatches(cacheMatches)
        except ValueError:
            cacheMatchesDict = cacheMatches

        return S_OK(cacheMatchesDict)

    def __getValidCache(self):
        """
        Obtains the keys on the cache which are valid. If any, returns the complete
        valid dictionary. If the list is empty, we assume the cache is invalid or
        not filled, so we issue a cache refresh and return its data.

        :return: { ( elementName, statusType, vO ) : status, ... }
        """

        cacheKeys = self.cacheKeys()
        # If cache is empty, we refresh it.
        if not cacheKeys:
            cache = self.refreshCache()
        else:
            cache = self.get(cacheKeys)

        return cache

    def __match(self, validCache, elementNames, elementType, statusTypes, vO):
        """
        Obtains all keys on the cache ( should not be empty ! ).

        Gets the sets ( no duplicates ) of elementNames and statusTypes. There is a
        slight distinction. A priori we cannot know which are all the elementNames.
        So, if elementNames is None, we will consider all elementNames in the cacheKeys.
        However, if statusTypes is None, we will get the standard list from the
        ResourceStatus configuration in the CS.

        If the cartesian product of our sets is on the cacheKeys set, we have a
        positive match.

        :Parameters:
          **validCache** - `dict`
            cache dictionary
          **elementNames** - [ None, `string`, `list` ]
            name(s) of the elements to be matched
          **elementType** - [ `string` ]
            type of the elements to be matched
          **statusTypes** - [ None, `string`, `list` ]
            name(s) of the statusTypes to be matched

        :return: S_OK() with a Vo check marker || S_ERROR()
        """

        cacheKeys = list(validCache)
        # flatten the cache. From our VO perspective we are only want to keep:
        # 1) keys with vO tuple element equal to our vO,
        # 2) keys with vO tuple element equal to 'all', but only if no element described in 1) exists.
        # a resource key is set to have 3 elements to allow a comparison with a cartesian product.
        checkVo = False
        if len(cacheKeys[0]) == 4:  # resource
            checkVo = True
            flattenedCache = {(key[0], key[1], key[2]): value for key, value in validCache.items() if key[3] == "all"}
            flattenedCache.update(
                {(key[0], key[1], key[2]): value for key, value in validCache.items() if key[3] == vO}
            )
            validCache = flattenedCache
        else:  # site, not VO specific in SiteStatus, eventually to be upgraded there to include the VO
            pass

        if isinstance(elementNames, six.string_types):
            elementNames = [elementNames]
        elif elementNames is None:
            if isinstance(cacheKeys[0], (tuple, list)):
                elementNames = [cacheKey[0] for cacheKey in cacheKeys]
            else:
                elementNames = cacheKeys
        # Remove duplicates, makes Cartesian product faster
        elementNamesSet = set(elementNames)

        if isinstance(elementType, six.string_types):
            if not elementType or elementType == "Site":
                elementType = []
            else:
                elementType = [elementType]
        elif elementType is None:
            elementType = [cacheKey[1] for cacheKey in cacheKeys]
        # Remove duplicates, makes Cartesian product faster
        elementTypeSet = set(elementType)

        if isinstance(statusTypes, six.string_types):
            if not statusTypes:
                statusTypes = []
            else:
                statusTypes = [statusTypes]
        elif statusTypes is None:
            statusTypes = self.allStatusTypes
        # Remove duplicates, makes Cartesian product faster
        statusTypesSet = set(statusTypes)

        if not elementTypeSet and not statusTypesSet:
            cartesianProduct = elementNamesSet
        else:
            cartesianProduct = set(itertools.product(elementNamesSet, elementTypeSet, statusTypesSet))

        # Some users find funny sending empty lists, which will make the cartesianProduct
        # be []. Problem: [] is always subset, no matter what !

        if not cartesianProduct:
            self.log.warn("Empty cartesian product")
            return S_ERROR("Empty cartesian product")

        notInCache = list(cartesianProduct.difference(set(validCache)))
        if notInCache:
            self.log.warn("Cache misses: %s" % notInCache)
            return S_ERROR("Cache misses: %s" % notInCache)

        result = S_OK(cartesianProduct)
        result["CheckVO"] = checkVo
        return result

    @staticmethod
    def __getDictFromCacheMatches(cacheMatches):
        """
        Formats the cacheMatches to a format expected by the RSS helpers clients.

        :Parameters:
          **cacheMatches** - `dict`
            cache dictionary of the form { ( elementName, elementType, statusType, vO ) : status, ... }


        :return: dict of the form { elementName : { statusType: status, ... }, ... }
        """

        result = {}

        for cacheKey, cacheValue in cacheMatches.items():
            elementName, _elementType, statusType, vO = cacheKey
            result.setdefault(elementName, {})[statusType] = cacheValue

        return result



""" Test class for SiteInspectorAgent
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# imports
from mock import MagicMock

# DIRAC Components
from DIRAC.ResourceStatusSystem.Agent.SiteInspectorAgent import SiteInspectorAgent
from DIRAC import gLogger

gLogger.setLevel("DEBUG")

# Mock Objects
mockAM = MagicMock()
mockNone = MagicMock()
mockNone.return_value = None
mockSM = MagicMock()

site = {
    "status": "status",
    "name": "site",
    "vO": "some_vo",
    "site": "site",
    "element": "Site",
    "statusType": "all",
    "elementType": "Site",
}


def test__execute(mocker):
    """Testing SiteInspectorAgent.execute()"""

    mocker.patch("DIRAC.ResourceStatusSystem.Agent.SiteInspectorAgent.AgentModule.__init__")
    mocker.patch("DIRAC.ResourceStatusSystem.Agent.SiteInspectorAgent.PEP")
    mocker.patch(
        "DIRAC.ResourceStatusSystem.Agent.SiteInspectorAgent.AgentModule._AgentModule__moduleProperties",
        side_effect=lambda x, y=None: y,
        create=True,
    )

    siteInspectorAgent = SiteInspectorAgent()
    siteInspectorAgent.log = gLogger
    siteInspectorAgent.log.setLevel("DEBUG")
    siteInspectorAgent._AgentModule__configDefaults = mockAM
    siteInspectorAgent.initialize()

    result = siteInspectorAgent._execute(site)

    assert result is None



"""
:mod: RssConfiguration

Module that collects utility functions.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK
from DIRAC.Core.Utilities import List
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.ResourceStatusSystem.PolicySystem.StateMachine import RSSMachine
from DIRAC.ResourceStatusSystem.Utilities import Utils

_rssConfigPath = "ResourceStatus"


class RssConfiguration(object):
    """
    RssConfiguration::

      {
        Config:
        {
          State        : Active | InActive,
          Cache        : 300,
          FromAddress  : 'email@site.domain'
          StatusType   :
          {
            default       : all,
            StorageElement: ReadAccess, WriteAccess, CheckAccess, RemoveAccess
          }
        }
      }

    """

    def __init__(self):
        self.opsHelper = Operations()

    def getConfigState(self, default="InActive"):
        """
        Gets from <pathToRSSConfiguration>/Config the value of State
        """

        return self.opsHelper.getValue("%s/Config/State" % _rssConfigPath, default)

    def getConfigCache(self, default=300):
        """
        Gets from <pathToRSSConfiguration>/Config the value of Cache
        """

        return self.opsHelper.getValue("%s/Config/Cache" % _rssConfigPath, default)

    def getConfigFromAddress(self, default=None):
        """
        Gets from <pathToRSSConfiguration>/Config the value of FromAddress
        """

        return self.opsHelper.getValue("%s/Config/FromAddress" % _rssConfigPath, default)

    def getConfigStatusType(self, elementType=None):
        """
        Gets all the status types per elementType, if not given, it takes default
        from CS. If not, hardcoded variable DEFAULT.
        """

        _DEFAULTS = ("all",)

        res = self.opsHelper.getOptionsDict("%s/Config/StatusTypes" % _rssConfigPath)

        if res["OK"]:

            if elementType in res["Value"]:
                return List.fromChar(res["Value"][elementType])

            if "default" in res["Value"]:
                return List.fromChar(res["Value"]["default"])

        return _DEFAULTS


def getPolicies():
    """
    Returns from the OperationsHelper: <_rssConfigPath>/Policies
    """

    return Utils.getCSTree("%s/Policies" % _rssConfigPath)


def getPolicyActions():
    """
    Returns from the OperationsHelper: <_rssConfigPath>/PolicyActions
    """

    return Utils.getCSTree("%s/PolicyActions" % _rssConfigPath)


def getnotificationGroups():
    """
    Returns from the OperationsHelper: <_rssConfigPath>/PolicyActions
    """

    return Utils.getCSTree("%s/Config" % _rssConfigPath)


def getNotifications():
    """
    Returns from the OperationsHelper: <_rssConfigPath>/Notification
    """

    return Utils.getCSTree("%s/Notification" % _rssConfigPath)


def getValidElements():
    """
    Returns from the OperationsHelper: <_rssConfigPath>/GeneralConfig/ValidElements
    """
    _DEFAULTS = ("Site", "Resource", "Node")

    #  result = Operations().getValue( '%s/GeneralConfig/ValidElements' % _rssConfigPath )
    #  if result is not None:
    #    return List.fromChar( result )
    return _DEFAULTS


def getValidStatus():
    """
    Returns a list of statuses as were defined on the RSS(State)Machine
    """

    validStatus = RSSMachine(None).getStates()
    return S_OK(validStatus)



"""
StdoutBackend wrapper
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import logging
import sys

from DIRAC.Resources.LogBackends.AbstractBackend import AbstractBackend
from DIRAC.FrameworkSystem.private.standardLogging.Formatter.ColoredBaseFormatter import ColoredBaseFormatter


class StdoutBackend(AbstractBackend):
    """
    StdoutBackend is used to create an abstraction of the handler and the formatter concepts from logging.
    Here, we gather a StreamHandler object and a BaseFormatter.

    - StreamHandler is from the standard logging library: it is used to write log messages in a desired stream
      so it needs a name: here it is stdout.

    - ColorBaseFormatter is a custom Formatter object, created for DIRAC in order to get the appropriate display
      with color.
      You can find it in FrameworkSystem/private/standardLogging/Formatter
    """

    def __init__(self, backendParams=None):
        super(StdoutBackend, self).__init__(logging.StreamHandler, ColoredBaseFormatter, backendParams)

    def _setHandlerParameters(self, backendParams=None):
        """
        Get the handler parameters from the backendParams.
        The keys of handlerParams should correspond to the parameter names of the associated handler.
        The method should be overridden in every backend that needs handler parameters.
        The method should be called before creating the handler object.

        :param dict parameters: parameters of the backend. ex: {'FileName': file.log}
        """
        self._handlerParams["stream"] = sys.stdout



""" DIRAC Multi VO FileCatalog plugin class to manage directory metadata for multiple VO.
"""
from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.DataManagementSystem.DB.FileCatalogComponents.DirectoryMetadata.DirectoryMetadata import DirectoryMetadata

VO_SUFFIX_SEPARATOR = "___"

# Metadata names mangling utilities


def _getMetaName(meta, credDict):
    """
    Return a fully-qualified metadata name based on client-supplied metadata name and
    client credentials. User VO is added to the metadata passed in.

    :param meta: metadata name
    :param credDict: client credentials
    :return: fully-qualified metadata name
    """

    return meta + _getMetaNameSuffix(credDict)


def _getMetaNameDict(metaDict, credDict):
    """
    Return a dictionary with fully-qualified metadata name keys based on client-supplied metadata name and
    client credentials. User VO is added to the metadata passed in.

    :param meta: metadata name
    :param credDict: client credentials
    :return: fully-qualified metadata name
    """

    fMetaDict = {}
    for meta, value in metaDict.items():
        fMetaDict[_getMetaName(meta, credDict)] = value
    return fMetaDict


def _getMetaNameSuffix(credDict):
    """
    Get a VO specific suffix from user credentials.

    :param credDict: user credentials
    :return: VO specific suffix
    """
    vo = Registry.getGroupOption(credDict["group"], "VO")
    return VO_SUFFIX_SEPARATOR + vo.replace("-", "_").replace(".", "_")


def _stripSuffix(metaDict, credDict):
    """
    Strip the suffix from all keys which contain it, removing all other keys.

    :param metaDict: original dict
    :param credDict: user credential dictionary
    :return: a new dict with modified keys
    """

    suffix = _getMetaNameSuffix(credDict)
    smetaDict = {key.rsplit(suffix, 1)[0]: value for key, value in metaDict.items() if key.endswith(suffix)}
    return smetaDict


class MultiVODirectoryMetadata(DirectoryMetadata):
    """
    Multi-VO FileCatalog plugin implementation.
    """

    def __init__(self, database=None):
        super(MultiVODirectoryMetadata, self).__init__(database=database)

    def addMetadataField(self, pName, pType, credDict):
        """
        Add a new metadata parameter to the Metadata Database.
        Modified to use fully qualified metadata names.

        :param str pName: parameter name
        :param str pType: parameter type in the MySQL notation
        :param dict credDict: client credential dictionary
        :return: standard Dirac result object
        """

        fname = _getMetaName(pName, credDict)
        return super(MultiVODirectoryMetadata, self).addMetadataField(fname, pType, credDict)

    def deleteMetadataField(self, pName, credDict):
        """Remove metadata field.
        Table name is now fully qualified

        :param str pName: parameter name
        :param dict credDict: client credential dictionary
        :return: standard Dirac result object
        """
        fname = _getMetaName(pName, credDict)
        return super(MultiVODirectoryMetadata, self).deleteMetadataField(fname, credDict)

    def getMetadataFields(self, credDict):
        """Get all the defined metadata fields

        :param dict credDict: client credential dictionary
        :return: standard Dirac result object
        """
        result = super(MultiVODirectoryMetadata, self)._getMetadataFields(credDict)
        if not result["OK"]:
            return result

        metaDict = _stripSuffix(result["Value"], credDict)
        return S_OK(metaDict)

    def setMetadata(self, dPath, metaDict, credDict):
        """Set the value of a given metadata field for the the given directory path

        :param str dPath: directory path
        :param dict metaDict: dictionary with the user metadata
        :param dict credDict: client credential dictionary

        :return: standard Dirac result object
        """
        fMetaDict = _getMetaNameDict(metaDict, credDict)
        return super(MultiVODirectoryMetadata, self).setMetadata(dPath, fMetaDict, credDict)

    def removeMetadata(self, dPath, metaList, credDict):
        """
        Remove the specified metadata for the given directory for users own VO.

        :param str dPath: directory path
        :param dict metaList: metadata names list
        :param dict credDict: client credential dictionary
        :return: standard Dirac result object
        """

        metaList = [_getMetaName(meta, credDict) for meta in metaList]
        result = super(MultiVODirectoryMetadata, self).removeMetadata(dPath, metaList, credDict)
        if not result["OK"]:
            if "FailedMetadata" in result:
                failedDict = _stripSuffix(result["FailedMetadata"], credDict)
                result["FailedMetadata"] = failedDict
            return result

        return S_OK()

    def setMetaParameter(self, dPath, metaName, metaValue, credDict):
        """
        Set an meta parameter - metadata which is not used in the the data
        search operations.

        :param str dPath: directory path
        :param str metaName: metadata name
        :param str metaValue: metadata value
        :param dict credDict: client credential dictionary
        :return: standard Dirac result object
        """
        fname = _getMetaName(metaName, credDict)
        return super(MultiVODirectoryMetadata, self).setMetaParameter(dPath, fname, metaValue, credDict)

    def getDirectoryMetadata(self, path, credDict, inherited=True, ownData=True):
        """
    Get metadata for the given directory aggregating metadata for the directory itself
    and for all the parent directories if inherited flag is True. Get also the non-indexed
    metadata parameters.

    :param str path: directory path
    :param dict credDict: client credential dictionary
    :param bool inherited: include parent directories if True
    :param bool ownData:
    :return: standard Dirac result object + additional MetadataOwner \
    and MetadataType dict entries if the operation is successful.
    """

        result = super(MultiVODirectoryMetadata, self).getDirectoryMetadata(path, credDict, inherited, ownData)
        if not result["OK"]:
            return result

        # Strip off the VO suffix
        result["Value"] = _stripSuffix(result["Value"], credDict)
        result["MetadataOwner"] = _stripSuffix(result["MetadataOwner"], credDict)
        result["MetadataType"] = _stripSuffix(result["MetadataType"], credDict)

        return result

    def findDirIDsByMetadata(self, metaDict, dPath, credDict):
        """Find Directories satisfying the given metadata and being subdirectories of
        the given path
        """
        fMetaDict = _getMetaNameDict(metaDict, credDict)
        return super(MultiVODirectoryMetadata, self).findDirIDsByMetadata(fMetaDict, dPath, credDict)



""" This agent syncs CS and pilot files to a web server of your choice

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN PilotSyncAgent
  :end-before: ##END
  :dedent: 2
  :caption: PilotsSyncAgent options

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import json
import shutil
import hashlib
import requests
import six

from DIRAC import S_OK
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Security.Locations import getHostCertificateAndKeyLocation, getCAsLocation
from DIRAC.DataManagementSystem.Client.DataManager import DataManager
from DIRAC.WorkloadManagementSystem.Utilities.PilotCStoJSONSynchronizer import PilotCStoJSONSynchronizer


class PilotSyncAgent(AgentModule):
    """Syncs CS and pilot files to a web server of your choice"""

    def __init__(self, *args, **kwargs):
        """c'tor"""
        super(PilotSyncAgent, self).__init__(*args, **kwargs)

        # This location would be enough if we are running this agent on the DIRAC web server
        # '/opt/dirac/webRoot/www/pilot'
        self.saveDir = ""
        self.uploadLocations = []
        self.includeMasterCS = True

    def initialize(self):
        """Initial settings"""
        self.workingDirectory = self.am_getOption("WorkDirectory")
        self.saveDir = self.am_getOption("SaveDirectory", self.saveDir)
        self.uploadLocations = self.am_getOption("UploadLocations", self.uploadLocations)
        includeMasterCS = self.am_getOption("IncludeMasterCS", self.includeMasterCS)
        if isinstance(includeMasterCS, six.string_types) and includeMasterCS.lower() in ["n", "no", "false"]:
            self.includeMasterCS = False

        self.certAndKeyLocation = getHostCertificateAndKeyLocation()
        self.casLocation = getCAsLocation()

        return S_OK()

    def execute(self):
        """cycle"""

        ps = PilotCStoJSONSynchronizer()
        ps.workDir = self.workingDirectory

        self.log.verbose("Parameters for this sync:")
        self.log.verbose("repo=" + ps.pilotRepo)
        self.log.verbose("VO repo=" + ps.pilotVORepo)
        self.log.verbose("projectDir=" + ps.projectDir)
        self.log.verbose("pilotScriptsPath=" + ps.pilotScriptPath)
        self.log.verbose("pilotVOScriptsPath=" + ps.pilotVOScriptPath)
        self.log.verbose("pilotRepoBranch=" + ps.pilotRepoBranch)
        self.log.verbose("pilotVORepoBranch=" + ps.pilotVORepoBranch)

        # pilot.json
        res = ps.getCSDict(includeMasterCS=self.includeMasterCS)
        if not res["OK"]:
            return res
        pilotDict = res["Value"]
        print(json.dumps(pilotDict, indent=4, sort_keys=True))  # just print here as formatting is important
        with open(os.path.join(self.workingDirectory, "pilot.json"), "w") as jf:
            json.dump(pilotDict, jf)

        # pilot files
        res = ps.syncScripts()
        if not res["OK"]:
            return res
        tarPath, tarFiles = res["Value"]

        allFiles = [tarPath] + tarFiles + [os.path.join(self.workingDirectory, "pilot.json")]

        # checksums
        checksumDict = {}
        for pFile in allFiles:
            filename = os.path.basename(pFile)
            with open(pFile, "rb") as fp:
                checksumDict[filename] = hashlib.sha512(fp.read()).hexdigest()
            cksPath = os.path.join(self.workingDirectory, "checksums.sha512")
        with open(cksPath, "wt") as chksums:
            for filename, chksum in sorted(checksumDict.items()):
                # same as the output from sha512sum commands
                chksums.write("%s  %s\n" % (chksum, filename))

        allFiles = list(set(allFiles + [cksPath]))

        if self.saveDir:
            # Moving files to the correct location
            self.log.info("Moving pilot files", "to %s" % self.saveDir)
            for tf in allFiles:
                # this overrides the destinations
                shutil.move(tf, os.path.join(self.saveDir, os.path.basename(tf)))

        # Here, attempting upload somewhere, and somehow
        for server in self.uploadLocations:
            self.log.info("Attempting to upload", "to %s" % server)
            if server.startswith("https://"):
                for tf in allFiles:
                    res = requests.put(server, data=tf, verify=self.casLocation, cert=self.certAndKeyLocation)
                    if res.status_code not in (200, 202):
                        self.log.error("Could not upload", "to %s: status %s" % (server, res.status_code))
            else:  # Assumes this is a DIRAC SE
                for tf in allFiles:
                    res = DataManager().put(lfn=tf, fileName=tf, diracSE=server)
                    if not res["OK"]:
                        self.log.error("Could not upload", "to %s: %s" % (server, res["Message"]))

        return S_OK()



""" Test_RSS_Policy_AlwaysActivePolicy
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id: $"

import unittest

import DIRAC.ResourceStatusSystem.Policy.AlwaysActivePolicy as moduleTested


################################################################################


class AlwaysActivePolicy_TestCase(unittest.TestCase):
    def setUp(self):
        """Setup"""

        self.moduleTested = moduleTested
        self.testClass = self.moduleTested.AlwaysActivePolicy

    def tearDown(self):
        """TearDown"""
        del self.testClass
        del self.moduleTested


################################################################################
# Tests


class AlwaysActivePolicy_Success(AlwaysActivePolicy_TestCase):
    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class"""

        policy = self.testClass()
        self.assertEqual("AlwaysActivePolicy", policy.__class__.__name__)

    def test_evaluate(self):
        """tests the evaluate method"""

        policy = self.testClass()
        res = policy.evaluate()

        self.assertTrue(res["OK"])
        self.assertEqual("Active", res["Value"]["Status"])


################################################################################
################################################################################

if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(AlwaysActivePolicy_TestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(AlwaysActivePolicy_Success))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



""" The Notification service provides a toolkit to contact people via email
    (eventually SMS etc.) to trigger some actions.

    The original motivation for this is due to some sites restricting the
    sending of email but it is useful for e.g. crash reports to get to their
    destination.

    Another use-case is for users to request an email notification for the
    completion of their jobs.  When output data files are uploaded to the
    Grid, an email could be sent by default with the metadata of the file.

    It can also be used to set alarms to be promptly forwarded to those
    subscribing to them.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
from DIRAC import gConfig, S_OK, S_ERROR

from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities.Mail import Mail
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler
from DIRAC.Core.Security import Properties
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.FrameworkSystem.DB.NotificationDB import NotificationDB
from DIRAC.Core.Utilities.DictCache import DictCache

__RCSID__ = "$Id$"


class NotificationHandler(RequestHandler):
    @classmethod
    def initializeHandler(cls, serviceInfo):
        """Handler initialization"""
        cls.notDB = NotificationDB()
        cls.mailCache = DictCache()
        gThreadScheduler.addPeriodicTask(3600, cls.notDB.purgeExpiredNotifications)
        gThreadScheduler.addPeriodicTask(3600, cls.mailCache.purgeExpired())
        return S_OK()

    def initialize(self):
        credDict = self.getRemoteCredentials()
        self.clientDN = credDict["DN"]
        self.clientGroup = credDict["group"]
        self.clientProperties = credDict["properties"]
        self.client = credDict["username"]

    ###########################################################################
    types_sendMail = [six.string_types, six.string_types, six.string_types, six.string_types]

    def export_sendMail(self, address, subject, body, fromAddress):
        """Send an email with supplied body to the specified address using the Mail utility.

        :param six.string_types address: recipient addresses
        :param six.string_types subject: subject of letter
        :param six.string_types body: body of letter
        :param six.string_types fromAddress: sender address, if None, will be used default from CS

        :return: S_OK(six.string_types)/S_ERROR() -- six.string_types is status message
        """
        self.log.verbose(
            "Received signal to send the following mail to %s:\nSubject = %s\n%s" % (address, subject, body)
        )
        if self.mailCache.exists(hash(address + subject + body)):
            return S_OK("Email with the same content already sent today to current addresses, come back tomorrow")
        eMail = Mail()
        notificationSection = PathFinder.getServiceSection("Framework/Notification")
        csSection = notificationSection + "/SMTP"
        eMail._smtpHost = gConfig.getValue("%s/Host" % csSection)
        eMail._smtpPort = gConfig.getValue("%s/Port" % csSection)
        eMail._smtpLogin = gConfig.getValue("%s/Login" % csSection)
        eMail._smtpPasswd = gConfig.getValue("%s/Password" % csSection)
        eMail._smtpPtcl = gConfig.getValue("%s/Protocol" % csSection)
        eMail._subject = subject
        eMail._message = body
        eMail._mailAddress = address
        if not fromAddress == "None":
            eMail._fromAddress = fromAddress
        eMail._fromAddress = gConfig.getValue("%s/FromAddress" % csSection) or eMail._fromAddress
        result = eMail._send()
        if not result["OK"]:
            self.log.warn("Could not send mail with the following message:\n%s" % result["Message"])
        else:
            self.mailCache.add(hash(address + subject + body), 3600 * 24)
            self.log.info("Mail sent successfully to %s with subject %s" % (address, subject))
            self.log.debug(result["Value"])

        return result

    ###########################################################################
    types_sendSMS = [six.string_types, six.string_types, six.string_types]

    def export_sendSMS(self, userName, body, fromAddress):
        """Send an SMS with supplied body to the specified DIRAC user using the Mail utility via an SMS switch.

        :param six.string_types userName: user name
        :param six.string_types body: message
        :param six.string_types fromAddress: sender address

        :return: S_OK()/S_ERROR()
        """
        self.log.verbose("Received signal to send the following SMS to %s:\n%s" % (userName, body))
        mobile = gConfig.getValue("/Registry/Users/%s/Mobile" % userName, "")
        if not mobile:
            return S_ERROR("No registered mobile number for %s" % userName)

        csSection = PathFinder.getServiceSection("Framework/Notification")
        smsSwitch = gConfig.getValue("%s/SMSSwitch" % csSection, "")
        if not smsSwitch:
            return S_ERROR("No SMS switch is defined in CS path %s/SMSSwitch" % csSection)

        address = "%s@%s" % (mobile, smsSwitch)
        subject = "DIRAC SMS"
        eMail = Mail()
        eMail._subject = subject
        eMail._message = body
        eMail._mailAddress = address
        if not fromAddress == "None":
            eMail._fromAddress = fromAddress
        result = eMail._send()
        if not result["OK"]:
            self.log.warn("Could not send SMS to %s with the following message:\n%s" % (userName, result["Message"]))
        else:
            self.log.info("SMS sent successfully to %s " % (userName))
            self.log.debug(result["Value"])

        return result

    ###########################################################################
    # ALARMS
    ###########################################################################

    types_newAlarm = [dict]

    def export_newAlarm(self, alarmDefinition):
        """Set a new alarm in the Notification database"""
        credDict = self.getRemoteCredentials()
        if "username" not in credDict:
            return S_ERROR("OOps. You don't have a username! This shouldn't happen :P")
        alarmDefinition["author"] = credDict["username"]
        return self.notDB.newAlarm(alarmDefinition)

    types_updateAlarm = [dict]

    def export_updateAlarm(self, updateDefinition):
        """update an existing alarm in the Notification database"""
        credDict = self.getRemoteCredentials()
        if "username" not in credDict:
            return S_ERROR("OOps. You don't have a username! This shouldn't happen :P")
        updateDefinition["author"] = credDict["username"]
        return self.notDB.updateAlarm(updateDefinition)

    types_getAlarmInfo = [six.integer_types]

    @classmethod
    def export_getAlarmInfo(cls, alarmId):
        """Get the extended info of an alarm"""
        result = cls.notDB.getAlarmInfo(alarmId)
        if not result["OK"]:
            return result
        alarmInfo = result["Value"]
        result = cls.notDB.getAlarmLog(alarmId)
        if not result["OK"]:
            return result
        return S_OK({"info": alarmInfo, "log": result["Value"]})

    types_getAlarms = [dict, list, int, int]

    @classmethod
    def export_getAlarms(cls, selectDict, sortList, startItem, maxItems):
        """Select existing alarms suitable for the Web monitoring"""
        return cls.notDB.getAlarms(selectDict, sortList, startItem, maxItems)

    types_deleteAlarmsByAlarmId = [(list, int)]

    @classmethod
    def export_deleteAlarmsByAlarmId(cls, alarmsIdList):
        """Delete alarms by alarmId"""
        return cls.notDB.deleteAlarmsByAlarmId(alarmsIdList)

    types_deleteAlarmsByAlarmKey = [(six.string_types, list)]

    @classmethod
    def export_deleteAlarmsByAlarmKey(cls, alarmsKeyList):
        """Delete alarms by alarmId"""
        return cls.notDB.deleteAlarmsByAlarmKey(alarmsKeyList)

    ###########################################################################
    # MANANGE ASSIGNEE GROUPS
    ###########################################################################

    types_setAssigneeGroup = [six.string_types, list]

    @classmethod
    def export_setAssigneeGroup(cls, groupName, userList):
        """Create a group of users to be used as an assignee for an alarm"""
        return cls.notDB.setAssigneeGroup(groupName, userList)

    types_getUsersInAssigneeGroup = [six.string_types]

    @classmethod
    def export_getUsersInAssigneeGroup(cls, groupName):
        """Get users in assignee group"""
        return cls.notDB.getUserAsignees(groupName)

    types_deleteAssigneeGroup = [six.string_types]

    @classmethod
    def export_deleteAssigneeGroup(cls, groupName):
        """Delete an assignee group"""
        return cls.notDB.deleteAssigneeGroup(groupName)

    types_getAssigneeGroups = []

    @classmethod
    def export_getAssigneeGroups(cls):
        """Get all assignee groups and the users that belong to them"""
        return cls.notDB.getAssigneeGroups()

    types_getAssigneeGroupsForUser = [six.string_types]

    def export_getAssigneeGroupsForUser(self, user):
        """Get all assignee groups and the users that belong to them"""
        credDict = self.getRemoteCredentials()
        if Properties.ALARMS_MANAGEMENT not in credDict["properties"]:
            user = credDict["username"]
        return self.notDB.getAssigneeGroupsForUser(user)

    ###########################################################################
    # MANAGE NOTIFICATIONS
    ###########################################################################

    types_addNotificationForUser = [six.string_types, six.string_types]

    def export_addNotificationForUser(self, user, message, lifetime=604800, deferToMail=True):
        """Create a group of users to be used as an assignee for an alarm"""
        try:
            lifetime = int(lifetime)
        except Exception:
            return S_ERROR("Message lifetime has to be a non decimal number")
        return self.notDB.addNotificationForUser(user, message, lifetime, deferToMail)

    types_removeNotificationsForUser = [six.string_types, list]

    def export_removeNotificationsForUser(self, user, notIds):
        """Get users in assignee group"""
        credDict = self.getRemoteCredentials()
        if Properties.ALARMS_MANAGEMENT not in credDict["properties"]:
            user = credDict["username"]
        return self.notDB.removeNotificationsForUser(user, notIds)

    types_markNotificationsAsRead = [six.string_types, list]

    def export_markNotificationsAsRead(self, user, notIds):
        """Delete an assignee group"""
        credDict = self.getRemoteCredentials()
        if Properties.ALARMS_MANAGEMENT not in credDict["properties"]:
            user = credDict["username"]
        return self.notDB.markNotificationsSeen(user, True, notIds)

    types_markNotificationsAsNotRead = [six.string_types, list]

    def export_markNotificationsAsNotRead(self, user, notIds):
        """Delete an assignee group"""
        credDict = self.getRemoteCredentials()
        if Properties.ALARMS_MANAGEMENT not in credDict["properties"]:
            user = credDict["username"]
        return self.notDB.markNotificationsSeen(user, False, notIds)

    types_getNotifications = [dict, list, int, int]

    def export_getNotifications(self, selectDict, sortList, startItem, maxItems):
        """Get all assignee groups and the users that belong to them"""
        credDict = self.getRemoteCredentials()
        if Properties.ALARMS_MANAGEMENT not in credDict["properties"]:
            selectDict["user"] = [credDict["username"]]
        return self.notDB.getNotifications(selectDict, sortList, startItem, maxItems)



"""  The Cloud Director is a simple agent performing VM instantiations
"""
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import six
import random
import socket
import hashlib
from collections import defaultdict

from DIRAC import S_OK, S_ERROR, gConfig
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.ConfigurationSystem.Client.Helpers import CSGlobals, Registry, Resources
from DIRAC.WorkloadManagementSystem.Client.MatcherClient import MatcherClient
from DIRAC.Core.Utilities.List import fromChar
from DIRAC.WorkloadManagementSystem.Client.ServerUtils import pilotAgentsDB
from DIRAC.ResourceStatusSystem.Client.SiteStatus import SiteStatus
from DIRAC.Resources.Cloud.EndpointFactory import EndpointFactory
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import (
    findGenericCloudCredentials,
    getVMTypes,
    getPilotBootstrapParameters,
)
from DIRAC.WorkloadManagementSystem.Client.ServerUtils import virtualMachineDB
from DIRAC.WorkloadManagementSystem.Utilities.Utils import getProxyFileForCloud

__RCSID__ = "$Id$"


class CloudDirector(AgentModule):
    """The CloudDirector works like a SiteDirector for cloud sites:
    It looks at the queued jobs in the task queues and attempts to
    start VM instances to meet the current demand.
    """

    def __init__(self, *args, **kwargs):
        super(CloudDirector, self).__init__(*args, **kwargs)
        self.vmTypeDict = {}
        self.vmTypeCECache = {}
        self.vmTypeSlots = {}
        self.failedVMTypes = defaultdict(int)
        self.firstPass = True

        self.vo = ""
        self.group = ""
        # self.voGroups contain all the eligible user groups for clouds submitted by this SiteDirector
        self.voGroups = []
        self.cloudDN = ""
        self.cloudGroup = ""
        self.platforms = []
        self.sites = []
        self.siteClient = None

        self.proxy = None

        self.updateStatus = True
        self.getOutput = False
        self.sendAccounting = True

    def initialize(self):
        self.siteClient = SiteStatus()
        return S_OK()

    def beginExecution(self):

        # The Director is for a particular user community
        self.vo = self.am_getOption("VO", "")
        if not self.vo:
            self.vo = CSGlobals.getVO()
        # The SiteDirector is for a particular user group
        self.group = self.am_getOption("Group", "")

        # Choose the group for which clouds will be submitted. This is a hack until
        # we will be able to match clouds to VOs.
        if not self.group:
            if self.vo:
                result = Registry.getGroupsForVO(self.vo)
                if not result["OK"]:
                    return result
                self.voGroups = []
                for group in result["Value"]:
                    if "NormalUser" in Registry.getPropertiesForGroup(group):
                        self.voGroups.append(group)
        else:
            self.voGroups = [self.group]

        result = findGenericCloudCredentials(vo=self.vo)
        if not result["OK"]:
            return result
        self.cloudDN, self.cloudGroup = result["Value"]
        self.maxVMsToSubmit = self.am_getOption("MaxVMsToSubmit", 1)
        self.runningPod = self.am_getOption("RunningPod", self.vo)

        # Get the site description dictionary
        siteNames = None
        if not self.am_getOption("Site", "Any").lower() == "any":
            siteNames = self.am_getOption("Site", [])
            if not siteNames:
                siteNames = None
        ces = None
        if not self.am_getOption("CEs", "Any").lower() == "any":
            ces = self.am_getOption("CEs", [])
            if not ces:
                ces = None

        result = getVMTypes(vo=self.vo, siteList=siteNames)
        if not result["OK"]:
            return result
        resourceDict = result["Value"]
        result = self.getEndpoints(resourceDict)
        if not result["OK"]:
            return result

        # if not siteNames:
        #  siteName = gConfig.getValue( '/DIRAC/Site', 'Unknown' )
        #  if siteName == 'Unknown':
        #    return S_OK( 'No site specified for the SiteDirector' )
        #  else:
        #    siteNames = [siteName]
        # self.siteNames = siteNames

        self.log.always("Sites:", siteNames)
        self.log.always("CEs:", ces)
        self.log.always("CloudDN:", self.cloudDN)
        self.log.always("CloudGroup:", self.cloudGroup)

        self.localhost = socket.getfqdn()
        self.proxy = ""

        if self.firstPass:
            if self.vmTypeDict:
                self.log.always("Agent will serve VM types:")
                for vmType in self.vmTypeDict:
                    self.log.always(
                        "Site: %s, CE: %s, VMType: %s"
                        % (self.vmTypeDict[vmType]["Site"], self.vmTypeDict[vmType]["CEName"], vmType)
                    )
        self.firstPass = False
        return S_OK()

    def __generateVMTypeHash(self, vmTypeDict):
        """Generate a hash of the queue description"""
        myMD5 = hashlib.md5()
        myMD5.update(str(sorted(vmTypeDict.items())).encode())
        hexstring = myMD5.hexdigest()
        return hexstring

    def getEndpoints(self, resourceDict):
        """Get the list of relevant CEs and their descriptions"""

        self.vmTypeDict = {}
        ceFactory = EndpointFactory()

        result = getPilotBootstrapParameters(vo=self.vo, runningPod=self.runningPod)
        if not result["OK"]:
            return result
        opParameters = result["Value"]

        for site in resourceDict:
            for ce in resourceDict[site]:
                ceDict = resourceDict[site][ce]
                ceTags = ceDict.get("Tag", [])
                if isinstance(ceTags, six.string_types):
                    ceTags = fromChar(ceTags)
                ceMaxRAM = ceDict.get("MaxRAM", None)
                qDict = ceDict.pop("VMTypes")
                for vmType in qDict:
                    vmTypeName = "%s_%s" % (ce, vmType)
                    self.vmTypeDict[vmTypeName] = {}
                    self.vmTypeDict[vmTypeName]["ParametersDict"] = qDict[vmType]
                    self.vmTypeDict[vmTypeName]["ParametersDict"]["VMType"] = vmType
                    self.vmTypeDict[vmTypeName]["ParametersDict"]["Site"] = site
                    self.vmTypeDict[vmTypeName]["ParametersDict"]["Setup"] = gConfig.getValue("/DIRAC/Setup", "unknown")
                    self.vmTypeDict[vmTypeName]["ParametersDict"]["CPUTime"] = 99999999

                    vmTypeTags = self.vmTypeDict[vmTypeName]["ParametersDict"].get("Tag")
                    if vmTypeTags and isinstance(vmTypeTags, six.string_types):
                        vmTypeTags = fromChar(vmTypeTags)
                        self.vmTypeDict[vmTypeName]["ParametersDict"]["Tag"] = vmTypeTags
                    if ceTags:
                        if vmTypeTags:
                            allTags = list(set(ceTags + vmTypeTags))
                            self.vmTypeDict[vmTypeName]["ParametersDict"]["Tag"] = allTags
                        else:
                            self.vmTypeDict[vmTypeName]["ParametersDict"]["Tag"] = ceTags

                    maxRAM = self.vmTypeDict[vmTypeName]["ParametersDict"].get("MaxRAM")
                    maxRAM = ceMaxRAM if not maxRAM else maxRAM
                    if maxRAM:
                        self.vmTypeDict[vmTypeName]["ParametersDict"]["MaxRAM"] = maxRAM

                    ceWholeNode = ceDict.get("WholeNode", "true")
                    wholeNode = self.vmTypeDict[vmTypeName]["ParametersDict"].get("WholeNode", ceWholeNode)
                    if wholeNode.lower() in ("yes", "true"):
                        self.vmTypeDict[vmTypeName]["ParametersDict"].setdefault("Tag", [])
                        self.vmTypeDict[vmTypeName]["ParametersDict"]["Tag"].append("WholeNode")

                    platform = ""
                    if "Platform" in self.vmTypeDict[vmTypeName]["ParametersDict"]:
                        platform = self.vmTypeDict[vmTypeName]["ParametersDict"]["Platform"]
                    elif "Platform" in ceDict:
                        platform = ceDict["Platform"]
                    if platform and platform not in self.platforms:
                        self.platforms.append(platform)

                    if "Platform" not in self.vmTypeDict[vmTypeName]["ParametersDict"] and platform:
                        result = Resources.getDIRACPlatform(platform)
                        if result["OK"]:
                            self.vmTypeDict[vmTypeName]["ParametersDict"]["Platform"] = result["Value"][0]

                    ceVMTypeDict = dict(ceDict)
                    ceVMTypeDict["CEName"] = ce
                    ceVMTypeDict["VO"] = self.vo
                    ceVMTypeDict["VMType"] = vmType
                    ceVMTypeDict["RunningPod"] = self.runningPod
                    ceVMTypeDict["CSServers"] = gConfig.getValue("/DIRAC/Configuration/Servers", [])
                    ceVMTypeDict.update(self.vmTypeDict[vmTypeName]["ParametersDict"])

                    # Allow a resource-specifc CAPath to be set (as some clouds have their own CAs)
                    # Otherwise fall back to the system-wide default(s)
                    if "CAPath" not in ceVMTypeDict:
                        ceVMTypeDict["CAPath"] = gConfig.getValue(
                            "/DIRAC/Security/CAPath", "/opt/dirac/etc/grid-security/certificates/cas.pem"
                        )

                    # Generate the CE object for the vmType or pick the already existing one
                    # if the vmType definition did not change
                    vmTypeHash = self.__generateVMTypeHash(ceVMTypeDict)
                    if vmTypeName in self.vmTypeCECache and self.vmTypeCECache[vmTypeName]["Hash"] == vmTypeHash:
                        vmTypeCE = self.vmTypeCECache[vmTypeName]["CE"]
                    else:
                        result = ceFactory.getCEObject(parameters=ceVMTypeDict)
                        if not result["OK"]:
                            return result
                        self.vmTypeCECache.setdefault(vmTypeName, {})
                        self.vmTypeCECache[vmTypeName]["Hash"] = vmTypeHash
                        self.vmTypeCECache[vmTypeName]["CE"] = result["Value"]
                        vmTypeCE = self.vmTypeCECache[vmTypeName]["CE"]
                        vmTypeCE.setBootstrapParameters(opParameters)

                    self.vmTypeDict[vmTypeName]["CE"] = vmTypeCE
                    self.vmTypeDict[vmTypeName]["CEName"] = ce
                    self.vmTypeDict[vmTypeName]["CEType"] = ceDict["CEType"]
                    self.vmTypeDict[vmTypeName]["Site"] = site
                    self.vmTypeDict[vmTypeName]["VMType"] = vmType
                    self.vmTypeDict[vmTypeName]["Platform"] = platform
                    self.vmTypeDict[vmTypeName]["MaxInstances"] = ceDict["MaxInstances"]
                    if not self.vmTypeDict[vmTypeName]["CE"].isValid():
                        self.log.error("Failed to instantiate CloudEndpoint for %s" % vmTypeName)
                        continue

                    if site not in self.sites:
                        self.sites.append(site)

        return S_OK()

    def execute(self):
        """Main execution method"""

        if not self.vmTypeDict:
            self.log.warn("No site defined, exiting the cycle")
            return S_OK()

        result = self.createVMs()
        if not result["OK"]:
            self.log.error("Errors in the job submission: ", result["Message"])

        # cyclesDone = self.am_getModuleParam( 'cyclesDone' )
        # if self.updateStatus and cyclesDone % self.cloudStatusUpdateCycleFactor == 0:
        #  result = self.updatePilotStatus()
        #  if not result['OK']:
        #    self.log.error( 'Errors in updating cloud status: ', result['Message'] )

        return S_OK()

    def createVMs(self):
        """Go through defined computing elements and submit jobs if necessary"""

        vmTypeList = list(self.vmTypeDict.keys())

        # Check that there is some work at all
        setup = CSGlobals.getSetup()
        tqDict = {"Setup": setup, "CPUTime": 9999999}
        if self.vo:
            tqDict["VO"] = self.vo
        if self.voGroups:
            tqDict["OwnerGroup"] = self.voGroups

        result = Resources.getCompatiblePlatforms(self.platforms)
        if not result["OK"]:
            return result
        tqDict["Platform"] = result["Value"]
        tqDict["Site"] = self.sites
        tags = []
        for vmType in vmTypeList:
            if "Tag" in self.vmTypeDict[vmType]["ParametersDict"]:
                tags += self.vmTypeDict[vmType]["ParametersDict"]["Tag"]
        tqDict["Tag"] = list(set(tags))

        self.log.verbose("Checking overall TQ availability with requirements")
        self.log.verbose(tqDict)

        matcherClient = MatcherClient()
        result = matcherClient.getMatchingTaskQueues(tqDict)
        if not result["OK"]:
            return result
        if not result["Value"]:
            self.log.verbose("No Waiting jobs suitable for the director")
            return S_OK()

        jobSites = set()
        anySite = False
        testSites = set()
        totalWaitingJobs = 0
        for tqID in result["Value"]:
            if "Sites" in result["Value"][tqID]:
                for site in result["Value"][tqID]["Sites"]:
                    if site.lower() != "any":
                        jobSites.add(site)
                    else:
                        anySite = True
            else:
                anySite = True
            if "JobTypes" in result["Value"][tqID]:
                if "Sites" in result["Value"][tqID]:
                    for site in result["Value"][tqID]["Sites"]:
                        if site.lower() != "any":
                            testSites.add(site)
            totalWaitingJobs += result["Value"][tqID]["Jobs"]

        tqIDList = list(result["Value"].keys())

        result = virtualMachineDB.getInstanceCounters("Status", {})
        totalVMs = 0
        if result["OK"]:
            for status in result["Value"]:
                if status in ["New", "Submitted", "Running"]:
                    totalVMs += result["Value"][status]
        self.log.info("Total %d jobs in %d task queues with %d VMs" % (totalWaitingJobs, len(tqIDList), totalVMs))

        # Check if the site is allowed in the mask
        result = self.siteClient.getUsableSites()
        if not result["OK"]:
            return S_ERROR("Can not get the site mask")
        siteMaskList = result.get("Value", [])

        vmTypeList = list(self.vmTypeDict.keys())
        random.shuffle(vmTypeList)
        totalSubmittedPilots = 0
        matchedQueues = 0
        for vmType in vmTypeList:
            ce = self.vmTypeDict[vmType]["CE"]
            ceName = self.vmTypeDict[vmType]["CEName"]
            vmTypeName = self.vmTypeDict[vmType]["VMType"]
            siteName = self.vmTypeDict[vmType]["Site"]
            platform = self.vmTypeDict[vmType]["Platform"]
            vmTypeTags = self.vmTypeDict[vmType]["ParametersDict"].get("Tag", [])
            siteMask = siteName in siteMaskList
            endpoint = "%s::%s" % (siteName, ceName)
            maxInstances = int(self.vmTypeDict[vmType]["MaxInstances"])
            processorTags = []

            # vms support WholeNode naturally
            processorTags.append("WholeNode")

            if not anySite and siteName not in jobSites:
                self.log.verbose("Skipping queue %s at %s: no workload expected" % (vmTypeName, siteName))
                continue
            if not siteMask and siteName not in testSites:
                self.log.verbose("Skipping queue %s: site %s not in the mask" % (vmTypeName, siteName))
                continue

            if "CPUTime" in self.vmTypeDict[vmType]["ParametersDict"]:
                vmTypeCPUTime = int(self.vmTypeDict[vmType]["ParametersDict"]["CPUTime"])
            else:
                self.log.warn("CPU time limit is not specified for queue %s, skipping..." % vmType)
                continue

            # Prepare the queue description to look for eligible jobs
            ceDict = ce.getParameterDict()

            if not siteMask:
                ceDict["JobType"] = "Test"
            if self.vo:
                ceDict["VO"] = self.vo
            if self.voGroups:
                ceDict["OwnerGroup"] = self.voGroups

            result = Resources.getCompatiblePlatforms(platform)
            if not result["OK"]:
                continue
            ceDict["Platform"] = result["Value"]

            ceDict["Tag"] = list(set(processorTags + vmTypeTags))

            # Get the number of eligible jobs for the target site/queue

            result = matcherClient.getMatchingTaskQueues(ceDict)
            if not result["OK"]:
                self.log.error("Could not retrieve TaskQueues from TaskQueueDB", result["Message"])
                return result
            taskQueueDict = result["Value"]
            if not taskQueueDict:
                self.log.verbose("No matching TQs found for %s" % vmType)
                continue

            matchedQueues += 1
            totalTQJobs = 0
            tqIDList = list(taskQueueDict.keys())
            for tq in taskQueueDict:
                totalTQJobs += taskQueueDict[tq]["Jobs"]

            self.log.verbose(
                "%d job(s) from %d task queue(s) are eligible for %s queue" % (totalTQJobs, len(tqIDList), vmType)
            )

            # Get the number of already instantiated VMs for these task queues
            totalWaitingVMs = 0
            result = virtualMachineDB.getInstanceCounters("Status", {"Endpoint": endpoint})
            if result["OK"]:
                for status in result["Value"]:
                    if status in ["New", "Submitted"]:
                        totalWaitingVMs += result["Value"][status]
            if totalWaitingVMs >= totalTQJobs:
                self.log.verbose("%d VMs already for all the available jobs" % totalWaitingVMs)

            self.log.verbose("%d VMs for the total of %d eligible jobs for %s" % (totalWaitingVMs, totalTQJobs, vmType))

            # Get proxy to be used to connect to the cloud endpoint
            authType = ce.parameters.get("Auth")
            if authType and authType.lower() in ["x509", "voms"]:
                self.log.verbose("Getting cloud proxy for %s/%s" % (siteName, ceName))
                result = getProxyFileForCloud(ce)
                if not result["OK"]:
                    continue
                ce.setProxy(result["Value"])

            # Get the number of available slots on the target site/endpoint
            totalSlots = self.getVMInstances(endpoint, maxInstances)
            if totalSlots == 0:
                self.log.debug("%s: No slots available" % vmType)
                continue

            vmsToSubmit = max(0, min(totalSlots, totalTQJobs - totalWaitingVMs))
            self.log.info(
                "%s: Slots=%d, TQ jobs=%d, VMs: %d, to submit=%d"
                % (vmType, totalSlots, totalTQJobs, totalWaitingVMs, vmsToSubmit)
            )

            # Limit the number of VM instances to create to vmsToSubmit
            vmsToSubmit = min(self.maxVMsToSubmit, vmsToSubmit)
            if vmsToSubmit == 0:
                continue

            self.log.info("Going to submit %d VMs to %s queue" % (vmsToSubmit, vmType))
            result = ce.createInstances(vmsToSubmit)

            # result = S_OK()
            if not result["OK"]:
                self.log.error("Failed submission to queue %s:\n" % vmType, result["Message"])
                self.failedVMTypes.setdefault(vmType, 0)
                self.failedVMTypes[vmType] += 1
                continue

            # Add VMs to the VirtualMachineDB
            vmDict = result["Value"]
            totalSubmittedPilots += len(vmDict)
            self.log.info("Submitted %d VMs to %s@%s" % (len(vmDict), vmTypeName, ceName))

            pilotList = []
            for uuID in vmDict:
                diracUUID = vmDict[uuID]["InstanceID"]
                endpoint = "%s::%s" % (self.vmTypeDict[vmType]["Site"], ceName)
                result = virtualMachineDB.insertInstance(uuID, vmTypeName, diracUUID, endpoint, self.vo)
                if not result["OK"]:
                    continue
                pRef = "vm://" + ceName + "/" + diracUUID + ":00"
                pilotList.append(pRef)

            stampDict = {}
            tqPriorityList = []
            sumPriority = 0.0
            for tq in taskQueueDict:
                sumPriority += taskQueueDict[tq]["Priority"]
                tqPriorityList.append((tq, sumPriority))
            tqDict = {}
            for pilotID in pilotList:
                rndm = random.random() * sumPriority
                for tq, prio in tqPriorityList:
                    if rndm < prio:
                        tqID = tq
                        break
                if tqID not in tqDict:
                    tqDict[tqID] = []
                tqDict[tqID].append(pilotID)

            for tqID, pilotList in tqDict.items():
                result = pilotAgentsDB.addPilotTQReference(pilotList, tqID, "", "", self.localhost, "Cloud", stampDict)
                if not result["OK"]:
                    self.log.error("Failed to insert pilots into the PilotAgentsDB: %s" % result["Message"])

        self.log.info(
            "%d VMs submitted in total in this cycle, %d matched queues" % (totalSubmittedPilots, matchedQueues)
        )
        return S_OK()

    def getVMInstances(self, endpoint, maxInstances):

        result = virtualMachineDB.getInstanceCounters("Status", {"Endpoint": endpoint})
        if not result["OK"]:
            return result

        count = 0
        for status in result["Value"]:
            if status in ["New", "Submitted", "Running"]:
                count += int(result["Value"][status])

        return max(0, maxInstances - count)



""" Utilities for WMS
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import io
import os
import sys
import json
import six

from DIRAC import gConfig, gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.File import mkDir
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import findGenericCloudCredentials


def createJobWrapper(
    jobID,
    jobParams,
    resourceParams,
    optimizerParams,
    extraOptions="",
    defaultWrapperLocation="DIRAC/WorkloadManagementSystem/JobWrapper/JobWrapperTemplate.py",
    log=gLogger,
    logLevel="INFO",
):
    """This method creates a job wrapper filled with the CE and Job parameters to execute the job.
    Main user is the JobAgent
    """
    if isinstance(extraOptions, six.string_types) and extraOptions.endswith(".cfg"):
        extraOptions = "--cfg %s" % extraOptions

    arguments = {"Job": jobParams, "CE": resourceParams, "Optimizer": optimizerParams}
    log.verbose("Job arguments are: \n %s" % (arguments))

    siteRoot = gConfig.getValue("/LocalSite/Root", os.getcwd())
    log.debug("SiteRootPythonDir is:\n%s" % siteRoot)
    workingDir = gConfig.getValue("/LocalSite/WorkingDirectory", siteRoot)
    mkDir("%s/job/Wrapper" % (workingDir))

    diracRoot = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))

    jobWrapperFile = "%s/job/Wrapper/Wrapper_%s" % (workingDir, jobID)
    if os.path.exists(jobWrapperFile):
        log.verbose("Removing existing Job Wrapper for %s" % (jobID))
        os.remove(jobWrapperFile)
    with open(os.path.join(diracRoot, defaultWrapperLocation), "r") as fd:
        wrapperTemplate = fd.read()

    if "LogLevel" in jobParams:
        logLevel = jobParams["LogLevel"]
        log.info("Found Job LogLevel JDL parameter with value: %s" % (logLevel))
    else:
        log.info("Applying default LogLevel JDL parameter with value: %s" % (logLevel))

    dPython = sys.executable
    realPythonPath = os.path.realpath(dPython)
    log.debug("Real python path after resolving links is: ", realPythonPath)
    dPython = realPythonPath

    # Making real substitutions
    # wrapperTemplate = wrapperTemplate.replace( "@JOBARGS@", str( arguments ) )
    wrapperTemplate = wrapperTemplate.replace("@SITEPYTHON@", str(siteRoot))

    jobWrapperJsonFile = jobWrapperFile + ".json"
    with io.open(jobWrapperJsonFile, "w", encoding="utf8") as jsonFile:
        json.dump(six.text_type(arguments), jsonFile, ensure_ascii=False)

    with open(jobWrapperFile, "w") as wrapper:
        wrapper.write(wrapperTemplate)

    jobExeFile = "%s/job/Wrapper/Job%s" % (workingDir, jobID)
    jobFileContents = """#!/bin/sh
%s %s %s -o LogLevel=%s -o /DIRAC/Security/UseServerCertificate=no
""" % (
        dPython,
        jobWrapperFile,
        extraOptions,
        logLevel,
    )
    with open(jobExeFile, "w") as jobFile:
        jobFile.write(jobFileContents)

    return S_OK(jobExeFile)


def createRelocatedJobWrapper(
    wrapperPath,
    rootLocation,
    jobID,
    jobParams,
    resourceParams,
    optimizerParams,
    extraOptions="",
    defaultWrapperLocation="DIRAC/WorkloadManagementSystem/JobWrapper/JobWrapperTemplate.py",
    log=gLogger,
    logLevel="INFO",
):
    """This method creates a job wrapper for a specific job in wrapperPath,
    but assumes this has been reloated to rootLocation before running it.
    """
    if isinstance(extraOptions, six.string_types) and extraOptions.endswith(".cfg") and "--cfg" not in extraOptions:
        extraOptions = "--cfg %s" % extraOptions

    arguments = {"Job": jobParams, "CE": resourceParams, "Optimizer": optimizerParams}
    log.verbose("Job arguments are: \n %s" % (arguments))

    diracRoot = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))

    jobWrapperFile = os.path.join(wrapperPath, "Wrapper_%s" % jobID)
    if os.path.exists(jobWrapperFile):
        log.verbose("Removing existing Job Wrapper for %s" % (jobID))
        os.remove(jobWrapperFile)
    with open(os.path.join(diracRoot, defaultWrapperLocation), "r") as fd:
        wrapperTemplate = fd.read()

    if "LogLevel" in jobParams:
        logLevel = jobParams["LogLevel"]
        log.info("Found Job LogLevel JDL parameter with value: %s" % (logLevel))
    else:
        log.info("Applying default LogLevel JDL parameter with value: %s" % (logLevel))

    # Making real substitutions
    # wrapperTemplate = wrapperTemplate.replace( "@JOBARGS@", str( arguments ) )
    wrapperTemplate = wrapperTemplate.replace("@SITEPYTHON@", rootLocation)

    jobWrapperJsonFile = jobWrapperFile + ".json"
    with io.open(jobWrapperJsonFile, "w", encoding="utf8") as jsonFile:
        json.dump(six.text_type(arguments), jsonFile, ensure_ascii=False)

    with open(jobWrapperFile, "w") as wrapper:
        wrapper.write(wrapperTemplate)

    # The "real" location of the jobwrapper after it is started
    jobWrapperDirect = os.path.join(rootLocation, "Wrapper_%s" % jobID)
    jobExeFile = os.path.join(wrapperPath, "Job%s" % jobID)
    jobFileContents = """#!/bin/sh
python %s %s -o LogLevel=%s -o /DIRAC/Security/UseServerCertificate=no
""" % (
        jobWrapperDirect,
        extraOptions,
        logLevel,
    )
    with open(jobExeFile, "w") as jobFile:
        jobFile.write(jobFileContents)

    jobExeDirect = os.path.join(rootLocation, "Job%s" % jobID)
    return S_OK(jobExeDirect)


def getProxyFileForCloud(ce):
    """Get a file with the proxy to be used to connect to the
        given cloud endpoint

    :param ce: cloud endpoint object
    :return: S_OK/S_ERROR, value is the path to the proxy file
    """

    vo = ce.parameters.get("VO")
    cloudDN = None
    cloudGroup = None
    if vo:
        result = findGenericCloudCredentials(vo=vo)
        if not result["OK"]:
            return result
        cloudDN, cloudGroup = result["Value"]

    cloudUser = ce.parameters.get("GenericCloudUser")
    if cloudUser:
        result = Registry.getDNForUsername(cloudUser)
        if not result["OK"]:
            return result
        cloudDN = result["Value"][0]
    cloudGroup = ce.parameters.get("GenericCloudGroup", cloudGroup)

    if cloudDN and cloudGroup:
        result = gProxyManager.getPilotProxyFromDIRACGroup(cloudDN, cloudGroup, 3600)
        if not result["OK"]:
            return result
        proxy = result["Value"]
        result = gProxyManager.dumpProxyToFile(proxy)
        return result
    else:
        return S_ERROR("Could not find generic cloud credentials")



""" Runs few integrity checks

The following options can be set for the ValidateOutputDataAgent.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN ValidateOutputDataAgent
  :end-before: ##END
  :dedent: 2
  :caption: ValidateOutputDataAgent options
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import re
import ast

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.DataManagementSystem.Client.DataIntegrityClient import DataIntegrityClient
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.Resources.Catalog.FileCatalogClient import FileCatalogClient
from DIRAC.TransformationSystem.Client.TransformationClient import TransformationClient
from DIRAC.DataManagementSystem.Client.ConsistencyInspector import ConsistencyInspector

AGENT_NAME = "Transformation/ValidateOutputDataAgent"


class ValidateOutputDataAgent(AgentModule):
    def __init__(self, *args, **kwargs):
        """c'tor"""
        AgentModule.__init__(self, *args, **kwargs)

        self.consistencyInspector = ConsistencyInspector()
        self.integrityClient = DataIntegrityClient()
        self.fc = FileCatalog()
        self.transClient = TransformationClient()
        self.fileCatalogClient = FileCatalogClient()

        agentTSTypes = self.am_getOption("TransformationTypes", [])
        if agentTSTypes:
            self.transformationTypes = agentTSTypes
        else:
            self.transformationTypes = Operations().getValue(
                "Transformations/DataProcessing", ["MCSimulation", "Merge"]
            )

        self.directoryLocations = sorted(
            self.am_getOption("DirectoryLocations", ["TransformationDB", "MetadataCatalog"])
        )
        self.transfidmeta = self.am_getOption("TransfIDMeta", "TransformationID")
        self.enableFlag = True

    #############################################################################

    def initialize(self):
        """Sets defaults"""

        # This sets the Default Proxy to used as that defined under
        # /Operations/Shifter/DataManager
        # the shifterProxy option in the Configuration can be used to change this default.
        self.am_setOption("shifterProxy", "DataManager")

        gLogger.info("Will treat the following transformation types: %s" % str(self.transformationTypes))
        gLogger.info("Will search for directories in the following locations: %s" % str(self.directoryLocations))
        gLogger.info("Will use %s as metadata tag name for TransformationID" % self.transfidmeta)
        return S_OK()

    #############################################################################

    def execute(self):
        """The VerifyOutputData execution method"""
        self.enableFlag = self.am_getOption("EnableFlag", "True")
        if not self.enableFlag == "True":
            self.log.info("VerifyOutputData is disabled by configuration option 'EnableFlag'")
            return S_OK("Disabled via CS flag")

        gLogger.info("-" * 40)
        self.updateWaitingIntegrity()
        gLogger.info("-" * 40)

        res = self.transClient.getTransformations({"Status": "ValidatingOutput", "Type": self.transformationTypes})
        if not res["OK"]:
            gLogger.error("Failed to get ValidatingOutput transformations", res["Message"])
            return res
        transDicts = res["Value"]
        if not transDicts:
            gLogger.info("No transformations found in ValidatingOutput status")
            return S_OK()
        gLogger.info("Found %s transformations in ValidatingOutput status" % len(transDicts))
        for transDict in transDicts:
            transID = transDict["TransformationID"]
            res = self.checkTransformationIntegrity(int(transID))
            if not res["OK"]:
                gLogger.error("Failed to perform full integrity check for transformation %d" % transID)
            else:
                self.finalizeCheck(transID)
                gLogger.info("-" * 40)
        return S_OK()

    def updateWaitingIntegrity(self):
        """Get 'WaitingIntegrity' transformations, update to 'ValidatedOutput'"""
        gLogger.info("Looking for transformations in the WaitingIntegrity status to update")
        res = self.transClient.getTransformations({"Status": "WaitingIntegrity"})
        if not res["OK"]:
            gLogger.error("Failed to get WaitingIntegrity transformations", res["Message"])
            return res
        transDicts = res["Value"]
        if not transDicts:
            gLogger.info("No transformations found in WaitingIntegrity status")
            return S_OK()
        gLogger.info("Found %s transformations in WaitingIntegrity status" % len(transDicts))
        for transDict in transDicts:
            transID = transDict["TransformationID"]
            gLogger.info("-" * 40)
            res = self.integrityClient.getTransformationProblematics(int(transID))
            if not res["OK"]:
                gLogger.error("Failed to determine waiting problematics for transformation", res["Message"])
            elif not res["Value"]:
                res = self.transClient.setTransformationParameter(transID, "Status", "ValidatedOutput")
                if not res["OK"]:
                    gLogger.error("Failed to update status of transformation %s to ValidatedOutput" % (transID))
                else:
                    gLogger.info("Updated status of transformation %s to ValidatedOutput" % (transID))
            else:
                gLogger.info("%d problematic files for transformation %s were found" % (len(res["Value"]), transID))
        return

    #############################################################################
    #
    # Get the transformation directories for checking
    #

    def getTransformationDirectories(self, transID):
        """Get the directories for the supplied transformation from the transformation system"""
        directories = []
        if "TransformationDB" in self.directoryLocations:
            res = self.transClient.getTransformationParameters(transID, ["OutputDirectories"])
            if not res["OK"]:
                gLogger.error("Failed to obtain transformation directories", res["Message"])
                return res
            if not isinstance(res["Value"], list):
                transDirectories = ast.literal_eval(res["Value"])
            else:
                transDirectories = res["Value"]
            directories = self._addDirs(transID, transDirectories, directories)

        if "MetadataCatalog" in self.directoryLocations:
            res = self.fileCatalogClient.findDirectoriesByMetadata({self.transfidmeta: transID})
            if not res["OK"]:
                gLogger.error("Failed to obtain metadata catalog directories", res["Message"])
                return res
            transDirectories = res["Value"]
            directories = self._addDirs(transID, transDirectories, directories)
        if not directories:
            gLogger.info("No output directories found")
        directories = sorted(directories)
        return S_OK(directories)

    @staticmethod
    def _addDirs(transID, newDirs, existingDirs):
        for nDir in newDirs:
            transStr = str(transID).zfill(8)
            if re.search(transStr, nDir):
                if nDir not in existingDirs:
                    existingDirs.append(nDir)
        return existingDirs

    #############################################################################
    def checkTransformationIntegrity(self, transID):
        """This method contains the real work"""
        gLogger.info("-" * 40)
        gLogger.info("Checking the integrity of transformation %s" % transID)
        gLogger.info("-" * 40)

        res = self.getTransformationDirectories(transID)
        if not res["OK"]:
            return res
        directories = res["Value"]
        if not directories:
            return S_OK()

        ######################################################
        #
        # This check performs Catalog->SE for possible output directories
        #
        res = self.fc.exists(directories)
        if not res["OK"]:
            gLogger.error("Failed to check directory existence", res["Message"])
            return res
        for directory, error in res["Value"]["Failed"]:
            gLogger.error("Failed to determine existance of directory", "%s %s" % (directory, error))
        if res["Value"]["Failed"]:
            return S_ERROR("Failed to determine the existance of directories")
        directoryExists = res["Value"]["Successful"]
        for directory in sorted(directoryExists):
            if not directoryExists[directory]:
                continue
            iRes = self.consistencyInspector.catalogDirectoryToSE(directory)
            if not iRes["OK"]:
                gLogger.error(iRes["Message"])
                return iRes

        gLogger.info("-" * 40)
        gLogger.info("Completed integrity check for transformation %s" % transID)
        return S_OK()

    def finalizeCheck(self, transID):
        """Move to 'WaitingIntegrity' or 'ValidatedOutput'"""
        res = self.integrityClient.getTransformationProblematics(int(transID))

        if not res["OK"]:
            gLogger.error("Failed to determine whether there were associated problematic files", res["Message"])
            newStatus = ""
        elif res["Value"]:
            gLogger.info("%d problematic files for transformation %s were found" % (len(res["Value"]), transID))
            newStatus = "WaitingIntegrity"
        else:
            gLogger.info("No problematics were found for transformation %s" % transID)
            newStatus = "ValidatedOutput"
        if newStatus:
            res = self.transClient.setTransformationParameter(transID, "Status", newStatus)
            if not res["OK"]:
                gLogger.error("Failed to update status of transformation %s to %s" % (transID, newStatus))
            else:
                gLogger.info("Updated status of transformation %s to %s" % (transID, newStatus))
        gLogger.info("-" * 40)
        return S_OK()



""" Priority corrector for the group and in-group shares
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.WorkloadManagementSystem.private.correctors.BaseCorrector import BaseCorrector


class SharesCorrector(object):
    def __init__(self, opsHelper):
        if not opsHelper:
            opsHelper = Operations()
        self.__opsHelper = opsHelper
        self.__log = gLogger.getSubLogger("SharesCorrector")
        self.__shareCorrectors = {}
        self.__correctorsOrder = []
        self.__baseCS = "JobScheduling/ShareCorrections"

    def __getCSValue(self, path, defaultValue=""):
        return self.__opsHelper.getValue("%s/%s" % (self.__baseCS, path), defaultValue)

    def __getCorrectorClass(self, correctorName):
        baseImport = "WorkloadManagementSystem.private.correctors"
        fullCN = "%s.%sCorrector" % (baseImport, correctorName)
        result = ObjectLoader().getObjects(baseImport, ".*Corrector", parentClass=BaseCorrector)
        if not result["OK"]:
            return result
        data = result["Value"]
        if fullCN not in data:
            return S_ERROR("Can't find corrector %s" % fullCN)
        return S_OK(data[fullCN])

    def instantiateRequiredCorrectors(self):
        correctorsToStart = self.__getCSValue("ShareCorrectorsToStart", [])
        self.__correctorsOrder = correctorsToStart
        self.__log.info("Correctors requested: %s" % ", ".join(correctorsToStart))
        for corrector in self.__shareCorrectors:
            if corrector not in correctorsToStart:
                self.__log.info("Stopping corrector %s" % corrector)
                del self.__shareCorrectors[corrector]
        for corrector in correctorsToStart:
            if corrector not in self.__shareCorrectors:
                self.__log.info("Starting corrector %s" % corrector)
                result = self.__opsHelper.getSections("%s/%s" % (self.__baseCS, corrector))
                if not result["OK"]:
                    self.__log.error(
                        "Cannot get list of correctors to instantiate",
                        " for corrector type %s: %s" % (corrector, result["Message"]),
                    )
                    continue
                groupCorrectors = result["Value"]
                self.__shareCorrectors[corrector] = {}
                result = self.__getCorrectorClass(corrector)
                if not result["OK"]:
                    self.__log.error("Cannot instantiate corrector", "%s %s" % (corrector, result["Message"]))
                    continue
                correctorClass = result["Value"]
                for groupCor in groupCorrectors:
                    groupPath = "%s/%s/Group" % (corrector, groupCor)
                    groupToCorrect = self.__getCSValue(groupPath, "")
                    if groupToCorrect:
                        groupKey = "gr:%s" % groupToCorrect
                    else:
                        groupKey = "global"
                    self.__log.info(
                        "Instantiating group corrector %s (%s) of type %s" % (groupCor, groupToCorrect, corrector)
                    )
                    if groupKey in self.__shareCorrectors[corrector]:
                        self.__log.error(
                            "There are two group correctors defined",
                            " for %s type (group %s)" % (corrector, groupToCorrect),
                        )
                    else:
                        groupCorPath = "%s/%s/%s" % (self.__baseCS, corrector, groupCor)
                        correctorObj = correctorClass(self.__opsHelper, groupCorPath, groupToCorrect)
                        result = correctorObj.initialize()
                        if not result["OK"]:
                            self.__log.error(
                                "Could not initialize corrector %s for %s: %s"
                                % (corrector, groupKey, result["Message"])
                            )
                        else:
                            self.__shareCorrectors[corrector][groupKey] = correctorObj
        return S_OK()

    def updateCorrectorsKnowledge(self):
        for corrector in self.__shareCorrectors:
            for groupTC in self.__shareCorrectors[corrector]:
                self.__shareCorrectors[corrector][groupTC].updateHistoryKnowledge()

    def update(self):
        self.instantiateRequiredCorrectors()
        self.updateCorrectorsKnowledge()

    def correctShares(self, shareDict, group=""):
        if group:
            groupKey = "gr:%s" % group
        else:
            groupKey = "global"
        for corrector in self.__shareCorrectors:
            if groupKey in self.__shareCorrectors[corrector]:
                shareDict = self.__shareCorrectors[corrector][groupKey].applyCorrection(shareDict)
        return shareDict



""" A set of utilities used in the WMS services
    Requires the Nordugrid ARC plugins. In particular : nordugrid-arc-python
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from tempfile import mkdtemp
import shutil

from DIRAC import S_OK, S_ERROR, gLogger, gConfig
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getQueue
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getGroupOption
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.Resources.Computing.ComputingElementFactory import ComputingElementFactory


# List of files to be inserted/retrieved into/from pilot Output Sandbox
# first will be defined as StdOut in JDL and the second as StdErr
outputSandboxFiles = ["StdOut", "StdErr"]

COMMAND_TIMEOUT = 60
###########################################################################


def getGridEnv():

    gridEnv = ""
    setup = gConfig.getValue("/DIRAC/Setup", "")
    if setup:
        instance = gConfig.getValue("/DIRAC/Setups/%s/WorkloadManagement" % setup, "")
        if instance:
            gridEnv = gConfig.getValue("/Systems/WorkloadManagement/%s/GridEnv" % instance, "")

    return gridEnv


def getPilotCE(pilotDict):
    """Instantiate and return a CE bound to a pilot"""
    ceFactory = ComputingElementFactory()
    result = getQueue(pilotDict["GridSite"], pilotDict["DestinationSite"], pilotDict["Queue"])
    if not result["OK"]:
        return result
    queueDict = result["Value"]
    gridEnv = getGridEnv()
    queueDict["GridEnv"] = gridEnv
    queueDict["WorkingDirectory"] = mkdtemp()
    result = ceFactory.getCE(pilotDict["GridType"], pilotDict["DestinationSite"], queueDict)
    if not result["OK"]:
        shutil.rmtree(queueDict["WorkingDirectory"])
        return result
    ce = result["Value"]
    return S_OK(ce)


def getPilotProxy(pilotDict):
    """Get a proxy bound to a pilot"""
    owner = pilotDict["OwnerDN"]
    group = pilotDict["OwnerGroup"]

    groupVOMS = getGroupOption(group, "VOMSRole", group)
    result = gProxyManager.getPilotProxyFromVOMSGroup(owner, groupVOMS)
    if not result["OK"]:
        gLogger.error("Could not get proxy:", 'User "%s" Group "%s" : %s' % (owner, groupVOMS, result["Message"]))
        return S_ERROR("Failed to get the pilot's owner proxy")
    proxy = result["Value"]
    return S_OK(proxy)


def getPilotRef(pilotReference, pilotDict):
    """Add the pilotStamp to the pilotReference, if the pilotStamp is in the dictionary,
    otherwise return unchanged pilotReference.
    """
    pilotStamp = pilotDict["PilotStamp"]
    pRef = pilotReference
    if pilotStamp:
        pRef = pRef + ":::" + pilotStamp
    return S_OK(pRef)


def killPilotsInQueues(pilotRefDict):
    """kill pilots queue by queue

    :params dict pilotRefDict: a dict of pilots in queues
    """

    ceFactory = ComputingElementFactory()
    failed = []
    for key, pilotDict in pilotRefDict.items():

        owner, group, site, ce, queue = key.split("@@@")
        result = getQueue(site, ce, queue)
        if not result["OK"]:
            return result
        queueDict = result["Value"]
        gridType = pilotDict["GridType"]
        result = ceFactory.getCE(gridType, ce, queueDict)
        if not result["OK"]:
            return result
        ce = result["Value"]

        # FIXME: quite hacky. Should be either removed, or based on some flag
        if gridType in ["CREAM", "ARC", "Globus", "HTCondorCE"]:
            group = getGroupOption(group, "VOMSRole", group)
            ret = gProxyManager.getPilotProxyFromVOMSGroup(owner, group)
            if not ret["OK"]:
                gLogger.error("Could not get proxy:", 'User "%s" Group "%s" : %s' % (owner, group, ret["Message"]))
                return S_ERROR("Failed to get the pilot's owner proxy")
            proxy = ret["Value"]
            ce.setProxy(proxy)

        pilotList = pilotDict["PilotList"]
        result = ce.killJob(pilotList)
        if not result["OK"]:
            failed.extend(pilotList)

    return failed



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import socket

##############################################################
#
#   Class to wrap Fake socket as if it was a real one
#
##############################################################


class FakeSocket:
    def __getattr__(self, name):
        return getattr(self.sock, name)

    def __init__(self, sock, copies=0):
        self.iCopies = copies
        self.sock = sock

    def close(self):
        if self.iCopies == 0:
            self.sock.shutdown()
            self.sock.close()
        else:
            self.iCopies -= 1

    def makefile(self, mode, bufsize=None):
        self.iCopies += 1
        return socket._fileobject(self.sock, mode, bufsize)  # pylint: disable=no-member



"""
  Instantiate the global Configuration Object
  gConfig is used everywhere within DIRAC to access Configuration data
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.ConfigurationSystem.private.ConfigurationClient import ConfigurationClient

#: Global gConfig object of type :class:`~DIRAC.ConfigurationSystem.private.ConfigurationClient.ConfigurationClient`
gConfig = ConfigurationClient()


def getConfig():
    """
    :returns: gConfig
    :rtype: ~DIRAC.ConfigurationSystem.private.ConfigurationClient.ConfigurationClient
    """
    return gConfig



#!/usr/bin/env python
"""
Get the size of the given file or a list of files

Example:
  $ dirac-dms-data-size  /formation/user/v/vhamar/Example.txt
  ------------------------------
  Files          |      Size (GB)
  ------------------------------
  1              |            0.0
  ------------------------------
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    unit = "GB"
    Script.registerSwitch("u:", "Unit=", "   Unit to use [default %s] (MB,GB,TB,PB)" % unit)
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(("LocalFile: Path to local file containing LFNs", "LFN:       Logical File Name"))
    Script.registerArgument(["LFN:       Logical File Name"], mandatory=False)
    unprSwitches, args = Script.parseCommandLine(ignoreErrors=False)

    for switch in unprSwitches:
        if switch[0].lower() == "u" or switch[0].lower() == "unit":
            unit = switch[1]
    scaleDict = {
        "MB": 1000 * 1000.0,
        "GB": 1000 * 1000 * 1000.0,
        "TB": 1000 * 1000 * 1000 * 1000.0,
        "PB": 1000 * 1000 * 1000 * 1000 * 1000.0,
    }
    if unit not in scaleDict.keys():
        gLogger.error("Unit must be one of MB,GB,TB,PB")
        DIRAC.exit(2)
    scaleFactor = scaleDict[unit]

    lfns = []
    for inputFileName in args:
        if os.path.exists(inputFileName):
            inputFile = open(inputFileName, "r")
            string = inputFile.read()
            inputFile.close()
            lfns.extend([lfn.strip() for lfn in string.splitlines()])
        else:
            lfns.append(inputFileName)

    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

    res = FileCatalog().getFileSize(lfns)
    if not res["OK"]:
        gLogger.error("Failed to get size of data", res["Message"])
        DIRAC.exit(-2)
    for lfn, reason in res["Value"]["Failed"].items():
        gLogger.error("Failed to get size for %s" % lfn, reason)
    totalSize = 0
    totalFiles = 0
    for lfn, size in res["Value"]["Successful"].items():
        totalFiles += 1
        totalSize += size
    gLogger.notice("-" * 30)
    gLogger.notice("%s|%s" % ("Files".ljust(15), ("Size (%s)" % unit).rjust(15)))
    gLogger.notice("-" * 30)
    gLogger.notice("%s|%s" % (str(totalFiles).ljust(15), str("%.1f" % (totalSize / scaleFactor)).rjust(15)))
    gLogger.notice("-" * 30)
    DIRAC.exit(0)


if __name__ == "__main__":
    main()



""" The SLURM TimeLeft utility interrogates the SLURM batch system for the
    current CPU consumed, as well as its limit.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.TimeLeft import runCommand
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.ResourceUsage import ResourceUsage


class SLURMResourceUsage(ResourceUsage):
    """
    This is the SLURM plugin of the TimeLeft Utility
    """

    def __init__(self):
        """Standard constructor"""
        super(SLURMResourceUsage, self).__init__("SLURM", "SLURM_JOB_ID")

        self.log.verbose("JOB_ID=%s" % self.jobID)

    def getResourceUsage(self):
        """Returns S_OK with a dictionary containing the entries CPU, CPULimit,
        WallClock, WallClockLimit, and Unit for current slot.
        """
        # sacct displays accounting data for all jobs and job steps
        # -j is the given job, -o the information of interest, -X to get rid of intermediate steps
        # -n to remove the header, -P to make the output parseable (remove tabs, spaces, columns)
        # --delimiter to specify character that splits the fields
        cmd = "sacct -j %s -o JobID,CPUTimeRAW,AllocCPUS,ElapsedRaw,Timelimit -X -n -P --delimiter=," % (self.jobID)
        result = runCommand(cmd)
        if not result["OK"]:
            return result

        cpu = None
        cpuLimit = None
        wallClock = None
        wallClockLimit = None

        output = str(result["Value"]).split(",")
        if len(output) == 5:
            _, cpu, allocCPUs, wallClock, wallClockLimitFormatted = output
            # Timelimit is in a specific format and have to be converted in seconds
            # TimelimitRaw is in seconds but only available from Slurm 18.08...
            wallClockLimit = self._getFormattedTimeInSeconds(wallClockLimitFormatted)
            wallClock = float(wallClock)
            if wallClockLimit:
                cpuLimit = wallClockLimit * int(allocCPUs)
            cpu = float(cpu)

        # Slurm allocations are based on wallclock time, not cpu time.
        # We precise it in the 'Unit' field
        consumed = {
            "CPU": cpu,
            "CPULimit": cpuLimit,
            "WallClock": wallClock,
            "WallClockLimit": wallClockLimit,
            "Unit": "WallClock",
        }

        if None in consumed.values():
            missed = [key for key, val in consumed.items() if val is None]
            msg = "Could not determine parameter"
            self.log.warn("Could not determine parameter", ",".join(missed))
            self.log.debug("This is the stdout from the batch system call\n%s" % (result["Value"]))
            return S_ERROR(msg)

        self.log.debug("TimeLeft counters complete:", str(consumed))
        return S_OK(consumed)

    def _getFormattedTimeInSeconds(self, slurmTime):
        """Convert SLURM time format into seconds

        According to the SLURM documentation, the format can be:
        - MM:SS
        - HH:MM:SS
        - DD-HH:MM:SS

        :param str slurmTime: time in SLURM format
        """
        slurmTimeList = slurmTime.split("-")
        try:
            # If slurmTime limit does not contain days
            if len(slurmTimeList) == 1:
                day = 0
                timeLeft = slurmTimeList[0]
            # Else
            elif len(slurmTimeList) == 2:
                day, timeLeft = slurmTimeList
            else:
                self.log.warn('Problem parsing "%s"' % slurmTime)
                return None

            timeLeftList = timeLeft.split(":")
            if len(timeLeftList) == 2:
                hours = 0
                minutes, seconds = timeLeftList
            elif len(timeLeftList) == 3:
                hours, minutes, seconds = timeLeftList
            else:
                self.log.warn('Problem parsing "%s"' % slurmTime)
                return None

            return ((int(day) * 24 + int(hours)) * 60 + int(minutes)) * 60 + float(seconds)
        except ValueError:
            self.log.warn('Problem parsing "%s"' % slurmTime)
            return None
